# -*- coding: utf-8 -*-
import logging
import random
from datetime import datetime, timedelta, timezone

from nectar.instance import shared_blockchain_instance
from nectarbase import operations

from .account import Account
from .amount import Amount
from .asset import Asset
from .price import FilledOrder, Order, Price
from .utils import (
    addTzInfo,
    assets_from_string,
    formatTimeFromNow,
    formatTimeString,
)

REQUEST_MODULE = None
if not REQUEST_MODULE:
    try:
        import requests

        REQUEST_MODULE = "requests"
    except ImportError:
        REQUEST_MODULE = None
log = logging.getLogger(__name__)


class Market(dict):
    """This class allows to easily access Markets on the blockchain for trading, etc.

    :param Steem blockchain_instance: Steem instance
    :param Asset base: Base asset
    :param Asset quote: Quote asset
    :returns: Blockchain Market
    :rtype: dictionary with overloaded methods

    Instances of this class are dictionaries that come with additional
    methods (see below) that allow dealing with a market and its
    corresponding functions.

    This class tries to identify **two** assets as provided in the
    parameters in one of the following forms:

    * ``base`` and ``quote`` are valid assets (according to :class:`nectar.asset.Asset`)
    * ``base:quote`` separated with ``:``
    * ``base/quote`` separated with ``/``
    * ``base-quote`` separated with ``-``

    .. note:: Throughout this library, the ``quote`` symbol will be
              presented first (e.g. ``STEEM:SBD`` with ``STEEM`` being the
              quote), while the ``base`` only refers to a secondary asset
              for a trade. This means, if you call
              :func:`nectar.market.Market.sell` or
              :func:`nectar.market.Market.buy`, you will sell/buy **only
              quote** and obtain/pay **only base**.

    """

    def __init__(self, base=None, quote=None, blockchain_instance=None, **kwargs):
        """
        Init Market

            :param nectar.steem.Steem blockchain_instance: Steem instance
            :param nectar.asset.Asset base: Base asset
            :param nectar.asset.Asset quote: Quote asset
        """
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if quote is None and isinstance(base, str):
            quote_symbol, base_symbol = assets_from_string(base)
            quote = Asset(quote_symbol, blockchain_instance=self.blockchain)
            base = Asset(base_symbol, blockchain_instance=self.blockchain)
            super(Market, self).__init__(
                {"base": base, "quote": quote}, blockchain_instance=self.blockchain
            )
        elif base and quote:
            quote = Asset(quote, blockchain_instance=self.blockchain)
            base = Asset(base, blockchain_instance=self.blockchain)
            super(Market, self).__init__(
                {"base": base, "quote": quote}, blockchain_instance=self.blockchain
            )
        elif base is None and quote is None:
            quote = Asset(self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain)
            base = Asset(self.blockchain.token_symbol, blockchain_instance=self.blockchain)
            super(Market, self).__init__(
                {"base": base, "quote": quote}, blockchain_instance=self.blockchain
            )
        else:
            raise ValueError("Unknown Market config")

    def get_string(self, separator=":"):
        """Return a formated string that identifies the market, e.g. ``STEEM:SBD``

        :param str separator: The separator of the assets (defaults to ``:``)
        """
        return "%s%s%s" % (self["quote"]["symbol"], separator, self["base"]["symbol"])

    def __eq__(self, other):
        if isinstance(other, str):
            quote_symbol, base_symbol = assets_from_string(other)
            return (
                self["quote"]["symbol"] == quote_symbol and self["base"]["symbol"] == base_symbol
            ) or (self["quote"]["symbol"] == base_symbol and self["base"]["symbol"] == quote_symbol)
        elif isinstance(other, Market):
            return (
                self["quote"]["symbol"] == other["quote"]["symbol"]
                and self["base"]["symbol"] == other["base"]["symbol"]
            )

    def ticker(self, raw_data=False):
        """Returns the ticker for all markets.

        Output Parameters:

        * ``latest``: Price of the order last filled
        * ``lowest_ask``: Price of the lowest ask
        * ``highest_bid``: Price of the highest bid
        * ``sbd_volume``: Volume of SBD
        * ``steem_volume``: Volume of STEEM
        * ``hbd_volume``: Volume of HBD
        * ``hive_volume``: Volume of HIVE
        * ``percent_change``: 24h change percentage (in %)

        .. note::
            Market is HIVE:HBD and prices are HBD per HIVE!

        Sample Output:

        .. code-block:: js

             {
                'highest_bid': 0.30100226633322913,
                'latest': 0.0,
                'lowest_ask': 0.3249636958897082,
                'percent_change': 0.0,
                'sbd_volume': 108329611.0,
                'steem_volume': 355094043.0
            }

        """
        data = {}
        # Core Exchange rate
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        ticker = self.blockchain.rpc.get_ticker(api="market_history")

        if raw_data:
            return ticker

        data["highest_bid"] = Price(
            ticker["highest_bid"],
            base=self["base"],
            quote=self["quote"],
            blockchain_instance=self.blockchain,
        )
        data["latest"] = Price(
            ticker["latest"],
            quote=self["quote"],
            base=self["base"],
            blockchain_instance=self.blockchain,
        )
        data["lowest_ask"] = Price(
            ticker["lowest_ask"],
            base=self["base"],
            quote=self["quote"],
            blockchain_instance=self.blockchain,
        )
        data["percent_change"] = float(ticker["percent_change"])
        if "sbd_volume" in ticker:
            data["sbd_volume"] = Amount(ticker["sbd_volume"], blockchain_instance=self.blockchain)
        elif "hbd_volume" in ticker:
            data["hbd_volume"] = Amount(ticker["hbd_volume"], blockchain_instance=self.blockchain)
        if "steem_volume" in ticker:
            data["steem_volume"] = Amount(
                ticker["steem_volume"], blockchain_instance=self.blockchain
            )
        elif "hive_volume" in ticker:
            data["hive_volume"] = Amount(ticker["hive_volume"], blockchain_instance=self.blockchain)

        return data

    def volume24h(self, raw_data=False):
        """Returns the 24-hour volume for all markets, plus totals for primary currencies.

        Sample output:

        .. code-block:: js

            {
                "STEEM": 361666.63617,
                "SBD": 1087.0
            }

        """
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        volume = self.blockchain.rpc.get_volume(api="market_history")
        if raw_data:
            return volume
        if "sbd_volume" in volume and "steem_volume" in volume:
            return {
                self["base"]["symbol"]: Amount(
                    volume["sbd_volume"], blockchain_instance=self.blockchain
                ),
                self["quote"]["symbol"]: Amount(
                    volume["steem_volume"], blockchain_instance=self.blockchain
                ),
            }
        elif "hbd_volume" in volume and "hive_volume" in volume:
            return {
                self["base"]["symbol"]: Amount(
                    volume["hbd_volume"], blockchain_instance=self.blockchain
                ),
                self["quote"]["symbol"]: Amount(
                    volume["hive_volume"], blockchain_instance=self.blockchain
                ),
            }

    def orderbook(self, limit=25, raw_data=False):
        """Returns the order book for SBD/STEEM market.

        :param int limit: Limit the amount of orders (default: 25)

        Sample output (raw_data=False):

            .. code-block:: none

                {
                    'asks': [
                        380.510 STEEM 460.291 SBD @ 1.209669 SBD/STEEM,
                        53.785 STEEM 65.063 SBD @ 1.209687 SBD/STEEM
                    ],
                    'bids': [
                        0.292 STEEM 0.353 SBD @ 1.208904 SBD/STEEM,
                        8.498 STEEM 10.262 SBD @ 1.207578 SBD/STEEM
                    ],
                    'asks_date': [
                        datetime.datetime(2018, 4, 30, 21, 7, 24, tzinfo=<UTC>),
                        datetime.datetime(2018, 4, 30, 18, 12, 18, tzinfo=<UTC>)
                    ],
                    'bids_date': [
                        datetime.datetime(2018, 4, 30, 21, 1, 21, tzinfo=<UTC>),
                        datetime.datetime(2018, 4, 30, 20, 38, 21, tzinfo=<UTC>)
                    ]
                }

        Sample output (raw_data=True):

            .. code-block:: js

                {
                    'asks': [
                        {
                            'order_price': {'base': '8.000 STEEM', 'quote': '9.618 SBD'},
                            'real_price': '1.20225000000000004',
                            'steem': 4565,
                            'sbd': 5488,
                            'created': '2018-04-30T21:12:45'
                        }
                    ],
                    'bids': [
                        {
                            'order_price': {'base': '10.000 SBD', 'quote': '8.333 STEEM'},
                            'real_price': '1.20004800192007677',
                            'steem': 8333,
                            'sbd': 10000,
                            'created': '2018-04-30T20:29:33'
                        }
                    ]
                }

        .. note:: Each bid is an instance of
            class:`nectar.price.Order` and thus carries the keys
            ``base``, ``quote`` and ``price``. From those you can
            obtain the actual amounts for sale

        """
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        if self.blockchain.rpc.get_use_appbase():
            orders = self.blockchain.rpc.get_order_book({"limit": limit}, api="market_history")
        else:
            orders = self.blockchain.rpc.get_order_book(limit, api="database_api")
        if raw_data:
            return orders
        asks = list(
            [
                Order(
                    Amount(x["order_price"]["quote"], blockchain_instance=self.blockchain),
                    Amount(x["order_price"]["base"], blockchain_instance=self.blockchain),
                    blockchain_instance=self.blockchain,
                )
                for x in orders["asks"]
            ]
        )
        bids = list(
            [
                Order(
                    Amount(x["order_price"]["quote"], blockchain_instance=self.blockchain),
                    Amount(x["order_price"]["base"], blockchain_instance=self.blockchain),
                    blockchain_instance=self.blockchain,
                ).invert()
                for x in orders["bids"]
            ]
        )
        asks_date = list([formatTimeString(x["created"]) for x in orders["asks"]])
        bids_date = list([formatTimeString(x["created"]) for x in orders["bids"]])
        data = {"asks": asks, "bids": bids, "asks_date": asks_date, "bids_date": bids_date}
        return data

    def recent_trades(self, limit=25, raw_data=False):
        """Returns the order book for a given market. You may also
        specify "all" to get the orderbooks of all markets.

        :param int limit: Limit the amount of orders (default: 25)
        :param bool raw_data: when False, FilledOrder objects will be
            returned

        Sample output (raw_data=False):

            .. code-block:: none

                [
                    (2018-04-30 21:00:54+00:00) 0.267 STEEM 0.323 SBD @ 1.209738 SBD/STEEM,
                    (2018-04-30 20:59:30+00:00) 0.131 STEEM 0.159 SBD @ 1.213740 SBD/STEEM,
                    (2018-04-30 20:55:45+00:00) 0.093 STEEM 0.113 SBD @ 1.215054 SBD/STEEM,
                    (2018-04-30 20:55:30+00:00) 26.501 STEEM 32.058 SBD @ 1.209690 SBD/STEEM,
                    (2018-04-30 20:55:18+00:00) 2.108 STEEM 2.550 SBD @ 1.209677 SBD/STEEM,
                ]

        Sample output (raw_data=True):

            .. code-block:: js

                [
                    {'date': '2018-04-30T21:02:45', 'current_pays': '0.235 SBD', 'open_pays': '0.194 STEEM'},
                    {'date': '2018-04-30T21:02:03', 'current_pays': '24.494 SBD', 'open_pays': '20.248 STEEM'},
                    {'date': '2018-04-30T20:48:30', 'current_pays': '175.464 STEEM', 'open_pays': '211.955 SBD'},
                    {'date': '2018-04-30T20:48:30', 'current_pays': '0.999 STEEM', 'open_pays': '1.207 SBD'},
                    {'date': '2018-04-30T20:47:54', 'current_pays': '0.273 SBD', 'open_pays': '0.225 STEEM'},
                ]

        .. note:: Each bid is an instance of
            :class:`nectar.price.Order` and thus carries the keys
            ``base``, ``quote`` and ``price``. From those you can
            obtain the actual amounts for sale

        """
        self.blockchain.rpc.set_next_node_on_empty_reply(limit > 0)
        if self.blockchain.rpc.get_use_appbase():
            orders = self.blockchain.rpc.get_recent_trades({"limit": limit}, api="market_history")[
                "trades"
            ]
        else:
            orders = self.blockchain.rpc.get_recent_trades(limit, api="market_history")
        if raw_data:
            return orders
        filled_order = list([FilledOrder(x, blockchain_instance=self.blockchain) for x in orders])
        return filled_order

    def trade_history(self, start=None, stop=None, intervall=None, limit=25, raw_data=False):
        """Returns the trade history for the internal market

        This function allows to fetch a fixed number of trades at fixed
        intervall times to reduce the call duration time. E.g. it is possible to
        receive the trades from the last 7 days, by fetching 100 trades each 6 hours.

        When intervall is set to None, all trades are received between start and stop.
        This can take a while.

        :param datetime start: Start date
        :param datetime stop: Stop date
        :param timedelta intervall: Defines the intervall
        :param int limit: Defines how many trades are fetched at each intervall point
        :param bool raw_data: when True, the raw data are returned
        """
        utc = timezone.utc
        if not stop:
            stop = utc.localize(datetime.now())
        if not start:
            start = stop - timedelta(hours=1)
        start = addTzInfo(start)
        stop = addTzInfo(stop)
        current_start = start
        filled_order = []
        fo = self.trades(start=current_start, stop=stop, limit=limit, raw_data=raw_data)
        if intervall is None and len(fo) > 0:
            current_start = fo[-1]["date"]
            filled_order += fo
        elif intervall is not None:
            current_start += intervall
            filled_order += [fo]
        last_date = fo[-1]["date"]
        while len(fo) > 0 and last_date < stop:
            fo = self.trades(start=current_start, stop=stop, limit=limit, raw_data=raw_data)
            if len(fo) == 0 or fo[-1]["date"] == last_date:
                break
            last_date = fo[-1]["date"]
            if intervall is None:
                current_start = last_date
                filled_order += fo
            else:
                current_start += intervall
                filled_order += [fo]
        return filled_order

    def trades(self, limit=100, start=None, stop=None, raw_data=False):
        """Returns your trade history for a given market.

        :param int limit: Limit the amount of orders (default: 100)
        :param datetime start: start time
        :param datetime stop: stop time

        """
        # FIXME, this call should also return whether it was a buy or
        # sell
        utc = timezone.utc
        if not stop:
            stop = utc.localize(datetime.now())
        if not start:
            start = stop - timedelta(hours=24)
        start = addTzInfo(start)
        stop = addTzInfo(stop)
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            orders = self.blockchain.rpc.get_trade_history(
                {"start": formatTimeString(start), "end": formatTimeString(stop), "limit": limit},
                api="market_history",
            )["trades"]
        else:
            orders = self.blockchain.rpc.get_trade_history(
                formatTimeString(start), formatTimeString(stop), limit, api="market_history"
            )
        if raw_data:
            return orders
        filled_order = list([FilledOrder(x, blockchain_instance=self.blockchain) for x in orders])
        return filled_order

    def market_history_buckets(self):
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        ret = self.blockchain.rpc.get_market_history_buckets(api="market_history")
        if self.blockchain.rpc.get_use_appbase():
            return ret["bucket_sizes"]
        else:
            return ret

    def market_history(self, bucket_seconds=300, start_age=3600, end_age=0, raw_data=False):
        """Return the market history (filled orders).

        :param int bucket_seconds: Bucket size in seconds (see
            `returnMarketHistoryBuckets()`)
        :param int start_age: Age (in seconds) of the start of the
            window (default: 1h/3600)
        :param int end_age: Age (in seconds) of the end of the window
            (default: now/0)
        :param bool raw_data: (optional) returns raw data if set True

        Example:

            .. code-block:: js

                {
                    'close_sbd': 2493387,
                    'close_steem': 7743431,
                    'high_sbd': 1943872,
                    'high_steem': 5999610,
                    'id': '7.1.5252',
                    'low_sbd': 534928,
                    'low_steem': 1661266,
                    'open': '2016-07-08T11:25:00',
                    'open_sbd': 534928,
                    'open_steem': 1661266,
                    'sbd_volume': 9714435,
                    'seconds': 300,
                    'steem_volume': 30088443
                }

        """
        buckets = self.market_history_buckets()
        if bucket_seconds < 5 and bucket_seconds >= 0:
            bucket_seconds = buckets[bucket_seconds]
        else:
            if bucket_seconds not in buckets:
                raise ValueError("You need select the bucket_seconds from " + str(buckets))
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            history = self.blockchain.rpc.get_market_history(
                {
                    "bucket_seconds": bucket_seconds,
                    "start": formatTimeFromNow(-start_age - end_age),
                    "end": formatTimeFromNow(-end_age),
                },
                api="market_history",
            )["buckets"]
        else:
            history = self.blockchain.rpc.get_market_history(
                bucket_seconds,
                formatTimeFromNow(-start_age - end_age),
                formatTimeFromNow(-end_age),
                api="market_history",
            )
        if raw_data:
            return history
        new_history = []
        for h in history:
            if "open" in h and isinstance(h.get("open"), str):
                h["open"] = formatTimeString(h.get("open", "1970-01-01T00:00:00"))
            new_history.append(h)
        return new_history

    def accountopenorders(self, account=None, raw_data=False):
        """Returns open Orders

        :param Account account: Account name or instance of Account to show orders for in this market
        :param bool raw_data: (optional) returns raw data if set True,
            or a list of Order() instances if False (defaults to False)
        """
        if not account:
            if "default_account" in self.blockchain.config:
                account = self.blockchain.config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, full=True, blockchain_instance=self.blockchain)

        r = []
        # orders = account["limit_orders"]
        if not self.blockchain.is_connected():
            return None
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            orders = self.blockchain.rpc.find_limit_orders(
                {"account": account["name"]}, api="database"
            )["orders"]
        else:
            orders = self.blockchain.rpc.get_open_orders(account["name"])
        if raw_data:
            return orders
        for o in orders:
            order = {}
            order["order"] = Order(
                Amount(o["sell_price"]["base"], blockchain_instance=self.blockchain),
                Amount(o["sell_price"]["quote"], blockchain_instance=self.blockchain),
                blockchain_instance=self.blockchain,
            )
            order["orderid"] = o["orderid"]
            order["created"] = formatTimeString(o["created"])
            r.append(order)
        return r

    def buy(
        self,
        price,
        amount,
        expiration=None,
        killfill=False,
        account=None,
        orderid=None,
        returnOrderId=False,
    ):
        """Places a buy order in a given market

        :param float price: price denoted in ``base``/``quote``
        :param number amount: Amount of ``quote`` to buy
        :param number expiration: (optional) expiration time of the order in seconds (defaults to 7 days)
        :param bool killfill: flag that indicates if the order shall be killed if it is not filled (defaults to False)
        :param string account: Account name that executes that order
        :param string returnOrderId: If set to "head" or "irreversible" the call will wait for the tx to appear in
            the head/irreversible block and add the key "orderid" to the tx output

        Prices/Rates are denoted in 'base', i.e. the SBD_STEEM market
        is priced in STEEM per SBD.

        **Example:** in the SBD_STEEM market, a price of 300 means
        a SBD is worth 300 STEEM

        .. note::

            All prices returned are in the **reversed** orientation as the
            market. I.e. in the STEEM/SBD market, prices are SBD per STEEM.
            That way you can multiply prices with `1.05` to get a +5%.

        .. warning::

            Since buy orders are placed as
            limit-sell orders for the base asset,
            you may end up obtaining more of the
            buy asset than you placed the order
            for. Example:

                * You place and order to buy 10 SBD for 100 STEEM/SBD
                * This means that you actually place a sell order for 1000 STEEM in order to obtain **at least** 10 SBD
                * If an order on the market exists that sells SBD for cheaper, you will end up with more than 10 SBD
        """
        if not expiration:
            expiration = self.blockchain.config["order-expiration"]
        if not account:
            if "default_account" in self.blockchain.config:
                account = self.blockchain.config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, blockchain_instance=self.blockchain)

        if isinstance(price, Price):
            price = price.as_base(self["base"]["symbol"])

        if isinstance(amount, Amount):
            amount = Amount(amount, blockchain_instance=self.blockchain)
            if not amount["asset"]["symbol"] == self["quote"]["symbol"]:
                raise AssertionError(
                    "Price: {} does not match amount: {}".format(str(price), str(amount))
                )
        elif isinstance(amount, str):
            amount = Amount(amount, blockchain_instance=self.blockchain)
        else:
            amount = Amount(amount, self["quote"]["symbol"], blockchain_instance=self.blockchain)
        order = operations.Limit_order_create(
            **{
                "owner": account["name"],
                "orderid": orderid or random.getrandbits(32),
                "amount_to_sell": Amount(
                    float(amount) * float(price),
                    self["base"]["symbol"],
                    blockchain_instance=self.blockchain,
                ),
                "min_to_receive": Amount(
                    float(amount), self["quote"]["symbol"], blockchain_instance=self.blockchain
                ),
                "expiration": formatTimeFromNow(expiration),
                "fill_or_kill": killfill,
                "prefix": self.blockchain.prefix,
                "json_str": not bool(self.blockchain.config["use_condenser"]),
            }
        )

        if returnOrderId:
            # Make blocking broadcasts
            prevblocking = self.blockchain.blocking
            self.blockchain.blocking = returnOrderId

        tx = self.blockchain.finalizeOp(order, account["name"], "active")

        if returnOrderId:
            tx["orderid"] = tx["operation_results"][0][1]
            self.blockchain.blocking = prevblocking

        return tx

    def sell(
        self,
        price,
        amount,
        expiration=None,
        killfill=False,
        account=None,
        orderid=None,
        returnOrderId=False,
    ):
        """Places a sell order in a given market

        :param float price: price denoted in ``base``/``quote``
        :param number amount: Amount of ``quote`` to sell
        :param number expiration: (optional) expiration time of the order in seconds (defaults to 7 days)
        :param bool killfill: flag that indicates if the order shall be killed if it is not filled (defaults to False)
        :param string account: Account name that executes that order
        :param string returnOrderId: If set to "head" or "irreversible" the call will wait for the tx to appear in
            the head/irreversible block and add the key "orderid" to the tx output

        Prices/Rates are denoted in 'base', i.e. the SBD_STEEM market
        is priced in STEEM per SBD.

        **Example:** in the SBD_STEEM market, a price of 300 means
        a SBD is worth 300 STEEM

        .. note::

            All prices returned are in the **reversed** orientation as the
            market. I.e. in the STEEM/SBD market, prices are SBD per STEEM.
            That way you can multiply prices with `1.05` to get a +5%.
        """
        if not expiration:
            expiration = self.blockchain.config["order-expiration"]
        if not account:
            if "default_account" in self.blockchain.config:
                account = self.blockchain.config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, blockchain_instance=self.blockchain)
        if isinstance(price, Price):
            price = price.as_base(self["base"]["symbol"])

        if isinstance(amount, Amount):
            amount = Amount(amount, blockchain_instance=self.blockchain)
            if not amount["asset"]["symbol"] == self["quote"]["symbol"]:
                raise AssertionError(
                    "Price: {} does not match amount: {}".format(str(price), str(amount))
                )
        elif isinstance(amount, str):
            amount = Amount(amount, blockchain_instance=self.blockchain)
        else:
            amount = Amount(amount, self["quote"]["symbol"], blockchain_instance=self.blockchain)
        order = operations.Limit_order_create(
            **{
                "owner": account["name"],
                "orderid": orderid or random.getrandbits(32),
                "amount_to_sell": Amount(
                    float(amount), self["quote"]["symbol"], blockchain_instance=self.blockchain
                ),
                "min_to_receive": Amount(
                    float(amount) * float(price),
                    self["base"]["symbol"],
                    blockchain_instance=self.blockchain,
                ),
                "expiration": formatTimeFromNow(expiration),
                "fill_or_kill": killfill,
                "prefix": self.blockchain.prefix,
                "json_str": not bool(self.blockchain.config["use_condenser"]),
            }
        )
        if returnOrderId:
            # Make blocking broadcasts
            prevblocking = self.blockchain.blocking
            self.blockchain.blocking = returnOrderId

        tx = self.blockchain.finalizeOp(order, account["name"], "active")

        if returnOrderId:
            tx["orderid"] = tx["operation_results"][0][1]
            self.blockchain.blocking = prevblocking

        return tx

    def cancel(self, orderNumbers, account=None, **kwargs):
        """Cancels an order you have placed in a given market. Requires
        only the "orderNumbers".

        :param orderNumbers: A single order number or a list of order numbers
        :type orderNumbers: int, list
        """
        if not account:
            if "default_account" in self.blockchain.config:
                account = self.blockchain.config["default_account"]
        if not account:
            raise ValueError("You need to provide an account")
        account = Account(account, full=False, blockchain_instance=self.blockchain)

        if not isinstance(orderNumbers, (list, set, tuple)):
            orderNumbers = {orderNumbers}

        op = []
        for order in orderNumbers:
            op.append(
                operations.Limit_order_cancel(
                    **{"owner": account["name"], "orderid": order, "prefix": self.blockchain.prefix}
                )
            )
        return self.blockchain.finalizeOp(op, account["name"], "active", **kwargs)

    @staticmethod
    def _weighted_average(values, weights):
        """Calculates a weighted average"""
        if not (len(values) == len(weights) and len(weights) > 0):
            raise AssertionError("Length of both array must be the same and greater than zero!")
        return sum(x * y for x, y in zip(values, weights)) / sum(weights)

    @staticmethod
    def btc_usd_ticker(verbose=False):
        """Returns the BTC/USD price from bitfinex, gdax, kraken, okcoin and bitstamp. The mean price is
        weighted by the exchange volume.
        """
        prices = {}
        responses = []
        urls = [
            # "https://api.bitfinex.com/v1/pubticker/BTCUSD",
            # "https://api.gdax.com/products/BTC-USD/ticker",
            # "https://api.kraken.com/0/public/Ticker?pair=XBTUSD",
            # "https://www.okcoin.com/api/v1/ticker.do?symbol=btc_usd",
            # "https://www.bitstamp.net/api/v2/ticker/btcusd/",
            "https://api.coingecko.com/api/v3/simple/price?ids=bitcoin&vs_currencies=usd&include_24hr_vol=true",
        ]
        cnt = 0
        while len(prices) == 0 and cnt < 5:
            cnt += 1
            try:
                responses = list(requests.get(u, timeout=30) for u in urls)
            except Exception as e:
                log.debug(str(e))

            for r in [
                x
                for x in responses
                if hasattr(x, "status_code") and x.status_code == 200 and x.json()
            ]:
                try:
                    if "bitfinex" in r.url:
                        data = r.json()
                        prices["bitfinex"] = {
                            "price": float(data["last_price"]),
                            "volume": float(data["volume"]),
                        }
                    elif "gdax" in r.url:
                        data = r.json()
                        prices["gdax"] = {
                            "price": float(data["price"]),
                            "volume": float(data["volume"]),
                        }
                    elif "kraken" in r.url:
                        data = r.json()["result"]["XXBTZUSD"]["p"]
                        prices["kraken"] = {"price": float(data[0]), "volume": float(data[1])}
                    elif "okcoin" in r.url:
                        data = r.json()["ticker"]
                        prices["okcoin"] = {
                            "price": float(data["last"]),
                            "volume": float(data["vol"]),
                        }
                    elif "bitstamp" in r.url:
                        data = r.json()
                        prices["bitstamp"] = {
                            "price": float(data["last"]),
                            "volume": float(data["volume"]),
                        }
                    elif "coingecko" in r.url:
                        data = r.json()["bitcoin"]
                        if "usd_24h_vol" in data:
                            volume = float(data["usd_24h_vol"])
                        else:
                            volume = 1
                        prices["coingecko"] = {"price": float(data["usd"]), "volume": volume}
                except KeyError as e:
                    log.info(str(e))

        if verbose:
            print(prices)

        if len(prices) == 0:
            raise RuntimeError("Obtaining BTC/USD prices has failed from all sources.")

        # vwap
        return Market._weighted_average(
            [x["price"] for x in prices.values()], [x["volume"] for x in prices.values()]
        )

    @staticmethod
    def steem_btc_ticker():
        """Returns the STEEM/BTC price from bittrex, binance, huobi and upbit. The mean price is
        weighted by the exchange volume.
        """
        prices = {}
        responses = []
        urls = [
            # "https://poloniex.com/public?command=returnTicker",
            # "https://bittrex.com/api/v1.1/public/getmarketsummary?market=BTC-STEEM",
            # "https://api.binance.com/api/v1/ticker/24hr",
            # "https://api.huobi.pro/market/history/kline?period=1day&size=1&symbol=steembtc",
            # "https://crix-api.upbit.com/v1/crix/trades/ticks?code=CRIX.UPBIT.BTC-STEEM&count=1",
            "https://api.coingecko.com/api/v3/simple/price?ids=steem&vs_currencies=btc&include_24hr_vol=true",
        ]
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.71 Safari/537.36",
        }
        cnt = 0
        while len(prices) == 0 and cnt < 5:
            cnt += 1
            try:
                responses = list(requests.get(u, headers=headers, timeout=30) for u in urls)
            except Exception as e:
                log.debug(str(e))

            for r in [
                x
                for x in responses
                if hasattr(x, "status_code") and x.status_code == 200 and x.json()
            ]:
                try:
                    if "poloniex" in r.url:
                        data = r.json()["BTC_STEEM"]
                        prices["poloniex"] = {
                            "price": float(data["last"]),
                            "volume": float(data["baseVolume"]),
                        }
                    elif "bittrex" in r.url:
                        data = r.json()["result"][0]
                        price = (data["Bid"] + data["Ask"]) / 2
                        prices["bittrex"] = {"price": price, "volume": data["BaseVolume"]}
                    elif "binance" in r.url:
                        data = [x for x in r.json() if x["symbol"] == "STEEMBTC"][0]
                        prices["binance"] = {
                            "price": float(data["lastPrice"]),
                            "volume": float(data["quoteVolume"]),
                        }
                    elif "huobi" in r.url:
                        data = r.json()["data"][-1]
                        prices["huobi"] = {
                            "price": float(data["close"]),
                            "volume": float(data["vol"]),
                        }
                    elif "upbit" in r.url:
                        data = r.json()[-1]
                        prices["upbit"] = {
                            "price": float(data["tradePrice"]),
                            "volume": float(data["tradeVolume"]),
                        }
                    elif "coingecko" in r.url:
                        data = r.json()["steem"]
                        if "usd_24h_vol" in data:
                            volume = float(data["usd_24h_vol"])
                        else:
                            volume = 1
                        prices["coingecko"] = {"price": float(data["btc"]), "volume": volume}
                except KeyError as e:
                    log.info(str(e))

        if len(prices) == 0:
            raise RuntimeError("Obtaining STEEM/BTC prices has failed from all sources.")

        return Market._weighted_average(
            [x["price"] for x in prices.values()], [x["volume"] for x in prices.values()]
        )

    def steem_usd_implied(self):
        """Returns the current STEEM/USD market price"""
        return self.steem_btc_ticker() * self.btc_usd_ticker()

    @staticmethod
    def hive_btc_ticker():
        """Returns the HIVE/BTC price from bittrex and upbit. The mean price is
        weighted by the exchange volume.
        """
        prices = {}
        responses = []
        urls = [
            # "https://bittrex.com/api/v1.1/public/getmarketsummary?market=BTC-HIVE",
            # "https://api.binance.com/api/v1/ticker/24hr",
            # "https://api.probit.com/api/exchange/v1/ticker?market_ids=HIVE-USDT",
            "https://api.coingecko.com/api/v3/simple/price?ids=hive&vs_currencies=btc&include_24hr_vol=true",
        ]
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.71 Safari/537.36",
        }
        cnt = 0
        while len(prices) == 0 and cnt < 5:
            cnt += 1
            try:
                responses = list(requests.get(u, headers=headers, timeout=30) for u in urls)
            except Exception as e:
                log.debug(str(e))

            for r in [
                x
                for x in responses
                if hasattr(x, "status_code") and x.status_code == 200 and x.json()
            ]:
                try:
                    if "poloniex" in r.url:
                        data = r.json()["BTC_HIVE"]
                        prices["poloniex"] = {
                            "price": float(data["last"]),
                            "volume": float(data["baseVolume"]),
                        }
                    elif "bittrex" in r.url:
                        data = r.json()["result"][0]
                        price = (data["Bid"] + data["Ask"]) / 2
                        prices["bittrex"] = {"price": price, "volume": data["BaseVolume"]}
                    elif "binance" in r.url:
                        data = [x for x in r.json() if x["symbol"] == "HIVEBTC"][0]
                        prices["binance"] = {
                            "price": float(data["lastPrice"]),
                            "volume": float(data["quoteVolume"]),
                        }
                    elif "huobi" in r.url:
                        data = r.json()["data"][-1]
                        prices["huobi"] = {
                            "price": float(data["close"]),
                            "volume": float(data["vol"]),
                        }
                    elif "upbit" in r.url:
                        data = r.json()[-1]
                        prices["upbit"] = {
                            "price": float(data["tradePrice"]),
                            "volume": float(data["tradeVolume"]),
                        }
                    elif "probit" in r.url:
                        data = r.json()["data"]
                        prices["huobi"] = {
                            "price": float(data["last"]),
                            "volume": float(data["base_volume"]),
                        }
                    elif "coingecko" in r.url:
                        data = r.json()["hive"]
                        if "btc_24h_vol":
                            volume = float(data["btc_24h_vol"])
                        else:
                            volume = 1
                        prices["coingecko"] = {"price": float(data["btc"]), "volume": volume}
                except KeyError as e:
                    log.info(str(e))

        if len(prices) == 0:
            raise RuntimeError("Obtaining HIVE/BTC prices has failed from all sources.")

        return Market._weighted_average(
            [x["price"] for x in prices.values()], [x["volume"] for x in prices.values()]
        )

    def hive_usd_implied(self):
        """Returns the current HIVE/USD market price"""
        return self.hive_btc_ticker() * self.btc_usd_ticker()



# -*- coding: utf-8 -*-
import unittest

from parameterized import parameterized

from nectar import Steem
from nectar.instance import set_shared_steem_instance
from nectar.witness import Witness, Witnesses, WitnessesRankedByVote, WitnessesVotedByAccount

from .nodes import get_hive_nodes, get_steem_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(
            node=get_hive_nodes(),
            nobroadcast=True,
            unsigned=True,
            keys={"active": wif},
            num_retries=10,
        )
        cls.steemit = Steem(
            node=get_steem_nodes(),
            nobroadcast=True,
            unsigned=True,
            keys={"active": wif},
            num_retries=10,
        )
        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        set_shared_steem_instance(cls.bts)
        cls.bts.set_default_account("test")

    @parameterized.expand(
        [
            ("normal"),
            ("steemit"),
        ]
    )
    def test_feed_publish(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        bts.txbuffer.clear()
        w = Witness("gtg", steem_instance=bts)
        tx = w.feed_publish("4 %s" % bts.backed_token_symbol, "1 %s" % bts.token_symbol)
        self.assertEqual((tx["operations"][0][0]), "feed_publish")
        op = tx["operations"][0][1]
        self.assertIn("gtg", op["publisher"])

    @parameterized.expand(
        [
            ("normal"),
            ("steemit"),
        ]
    )
    def test_update(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        bts.txbuffer.clear()
        w = Witness("gtg", steem_instance=bts)
        props = {
            "account_creation_fee": "0.1 %s" % bts.token_symbol,
            "maximum_block_size": 32000,
            "sbd_interest_rate": 0,
        }
        tx = w.update(wif, "", props)
        self.assertEqual((tx["operations"][0][0]), "witness_update")
        op = tx["operations"][0][1]
        self.assertIn("gtg", op["owner"])

    @parameterized.expand(
        [
            ("normal"),
            ("steemit"),
        ]
    )
    def test_witnesses(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        w = Witnesses(steem_instance=bts)
        w.printAsTable()
        self.assertTrue(len(w) > 0)
        self.assertTrue(isinstance(w[0], Witness))

    @parameterized.expand(
        [
            ("normal"),
        ]
    )
    def test_WitnessesVotedByAccount(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        w = WitnessesVotedByAccount("gtg", steem_instance=bts)
        w.printAsTable()
        self.assertTrue(len(w) > 0)
        self.assertTrue(isinstance(w[0], Witness))

    @parameterized.expand(
        [
            ("normal"),
            ("steemit"),
        ]
    )
    def test_WitnessesRankedByVote(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        w = WitnessesRankedByVote(steem_instance=bts)
        w.printAsTable()
        self.assertTrue(len(w) > 0)
        self.assertTrue(isinstance(w[0], Witness))

    @parameterized.expand(
        [
            ("normal"),
            ("steemit"),
        ]
    )
    def test_export(self, node_param):
        if node_param == "normal":
            bts = self.bts
        else:
            bts = self.steemit
        owner = "gtg"
        if bts.rpc.get_use_appbase():
            witness = bts.rpc.find_witnesses({"owners": [owner]}, api="database")["witnesses"]
            if len(witness) > 0:
                witness = witness[0]
        else:
            witness = bts.rpc.get_witness_by_account(owner)

        w = Witness(owner, steem_instance=bts)
        keys = list(witness.keys())
        json_witness = w.json()
        exclude_list = [
            "votes",
            "virtual_last_update",
            "virtual_scheduled_time",
            "last_aslot",
            "last_confirmed_block_num",
            "available_witness_account_subsidies",
        ]
        for k in keys:
            if k not in exclude_list:
                if isinstance(witness[k], dict) and isinstance(json_witness[k], list):
                    self.assertEqual(list(witness[k].values()), json_witness[k])
                else:
                    self.assertEqual(witness[k], json_witness[k])



# This Python file uses the following encoding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import unittest

from nectargraphenebase import objects, types


class Testcases(unittest.TestCase):
    def test_GrapheneObject(self):
        j = {"a": 2, "b": "abcde", "c": ["a", "b"]}
        j2 = objects.GrapheneObject(j)
        self.assertEqual(j, j2.data)
        self.assertEqual(json.loads(j2.__str__()), j2.json())

        a = objects.Array(["1000", 3, "@@000000013"])
        j = {"a": a}
        j2 = objects.GrapheneObject(j)
        self.assertEqual(j, j2.data)
        self.assertEqual(json.loads(j2.__str__()), j2.json())

        a = types.Array(["1000", 3, "@@000000013"])
        j = {"a": a}
        j2 = objects.GrapheneObject(j)
        self.assertEqual(j, j2.data)
        self.assertEqual(json.loads(j2.__str__()), j2.json())



# -*- coding: utf-8 -*-
import unittest

import mock

from nectar import Hive
from nectar.account import Account
from nectar.instance import set_shared_steem_instance
from nectar.message import Message

from .nodes import get_hive_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"
core_unit = "STM"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Hive(node=get_hive_nodes(), nobroadcast=True, keys=[wif], num_retries=10)
        set_shared_steem_instance(cls.bts)

    def test_sign_message(self):
        def new_refresh(self):
            dict.__init__(
                self,
                {
                    "identifier": "test",
                    "name": "test",
                    "id_item": "name",
                    "memo_key": "STM6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV",
                },
            )

        with mock.patch("nectar.account.Account.refresh", new=new_refresh):
            account = Account("test")
            account["memo_key"] = "STM6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV"
            p = Message("message foobar").sign(account=account)
            Message(p).verify(account=account)

    def test_verify_message(self):
        def new_refresh(self):
            dict.__init__(
                self,
                {
                    "identifier": "test",
                    "name": "test",
                    "id_item": "name",
                    "memo_key": "STM6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV",
                },
            )

        with mock.patch("nectar.account.Account.refresh", new=new_refresh):
            Message(
                "-----BEGIN HIVE SIGNED MESSAGE-----\n"
                "message foobar\n"
                "-----BEGIN META-----\n"
                "account=test\n"
                "memokey=STM6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV\n"
                "block=19902522\n"
                "timestamp=2018-02-15T22:00:54\n"
                "-----BEGIN SIGNATURE-----\n"
                "20093ef63f375b9aa8570188cae3aad953bf6393d43ce6f03bbbd1b429e48c6a587dc012922515f6d327158df5081ea2d595888225f9f1c6c3028781c8f9451fde\n"
                "-----END HIVE SIGNED MESSAGE-----\n"
            ).verify()

            Message(
                "-----BEGIN HIVE SIGNED MESSAGE-----"
                "message foobar\n"
                "-----BEGIN META-----"
                "account=test\n"
                "memokey=STM6MRyAjQq8ud7hVNYcfnVPJqcVpscN5So8BhtHuGYqET5GDW5CV\n"
                "block=19902522\n"
                "timestamp=2018-02-15T22:00:54\n"
                "-----BEGIN SIGNATURE-----"
                "20093ef63f375b9aa8570188cae3aad953bf6393d43ce6f03bbbd1b429e48c6a587dc012922515f6d327158df5081ea2d595888225f9f1c6c3028781c8f9451fde\n"
                "-----END HIVE SIGNED MESSAGE-----"
            ).verify()



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem, exceptions
from nectar.account import Account, extract_account_name
from nectar.amount import Amount
from nectar.block import Block
from nectar.instance import set_shared_blockchain_instance
from nectar.utils import formatTimeString

from .nodes import get_steem_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(
            node=get_steem_nodes(),
            nobroadcast=True,
            bundle=False,
            unsigned=True,
            # Overwrite wallet to use this list of wifs only
            keys={"active": wif},
            num_retries=10,
        )
        cls.account = Account("thecrazygm", steem_instance=cls.bts)
        set_shared_blockchain_instance(cls.bts)

    def test_account(self):
        stm = self.bts
        account = self.account
        Account("thecrazygm", steem_instance=stm)
        with self.assertRaises(exceptions.AccountDoesNotExistsException):
            Account("DoesNotExistsXXX", steem_instance=stm)
        # asset = Asset("1.3.0")
        # symbol = asset["symbol"]
        self.assertEqual(account.name, "thecrazygm")
        self.assertEqual(account["name"], account.name)
        self.assertIsInstance(account.get_balance("available", "SBD"), Amount)
        account.print_info()
        # self.assertIsInstance(account.balance({"symbol": symbol}), Amount)
        self.assertIsInstance(account.available_balances, list)
        self.assertTrue(account.virtual_op_count() > 0)

        # BlockchainObjects method
        account.cached = False
        self.assertTrue(list(account.items()))
        account.cached = False
        self.assertIn("id", account)
        account.cached = False
        # self.assertEqual(account["id"], "1.2.1")
        self.assertEqual(str(account), "<Account thecrazygm>")
        self.assertIsInstance(Account(account), Account)

    def test_history(self):
        account = self.account
        zero_element = 0
        h_all_raw = []
        for h in account.history_reverse(raw_output=True):
            h_all_raw.append(h)
        # h_all_raw = h_all_raw[zero_element:]
        zero_element = h_all_raw[-1][0]
        h_list = []
        for h in account.history(stop=10, use_block_num=False, batch_size=10, raw_output=True):
            h_list.append(h)
        # self.assertEqual(h_list[0][0], zero_element)
        self.assertEqual(h_list[-1][0], 10)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-1][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-11 + zero_element][1]["block"])
        h_list = []
        for h in account.history(
            start=1, stop=9, use_block_num=False, batch_size=10, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 1)
        self.assertEqual(h_list[-1][0], 9)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        start = formatTimeString(h_list[0][1]["timestamp"])
        stop = formatTimeString(h_list[-1][1]["timestamp"])
        h_list = []
        for h in account.history(
            start=start, stop=stop, use_block_num=False, batch_size=10, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 1)
        self.assertEqual(h_list[-1][0], 9)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        h_list = []
        for h in account.history_reverse(
            start=10, stop=0, use_block_num=False, batch_size=10, raw_output=False
        ):
            h_list.append(h)
        # zero_element = h_list[-1]['index']
        self.assertEqual(h_list[0]["index"], 10)
        # self.assertEqual(h_list[-1]['index'], zero_element)
        self.assertEqual(h_list[0]["block"], h_all_raw[-11 + zero_element][1]["block"])
        self.assertEqual(h_list[-1]["block"], h_all_raw[-1][1]["block"])
        h_list = []
        for h in account.history_reverse(
            start=9, stop=1, use_block_num=False, batch_size=10, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 9)
        self.assertEqual(h_list[-1][0], 1)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        start = formatTimeString(h_list[0][1]["timestamp"])
        stop = formatTimeString(h_list[-1][1]["timestamp"])
        h_list = []
        for h in account.history_reverse(
            start=start, stop=stop, use_block_num=False, batch_size=10, raw_output=True
        ):
            h_list.append(h)
        # self.assertEqual(h_list[0][0], 8)
        self.assertEqual(h_list[-1][0], 1)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        h_list = []
        for h in account.get_account_history(10, 10, use_block_num=False, order=1, raw_output=True):
            h_list.append(h)
        # self.assertEqual(h_list[0][0], zero_element)
        self.assertEqual(h_list[-1][0], 10)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-1][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-11 + zero_element][1]["block"])
        h_list = []
        for h in account.get_account_history(
            10, 10, use_block_num=False, start=1, stop=9, order=1, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 1)
        self.assertEqual(h_list[-1][0], 9)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        start = formatTimeString(h_list[0][1]["timestamp"])
        stop = formatTimeString(h_list[-1][1]["timestamp"])
        h_list = []
        for h in account.get_account_history(
            10, 10, use_block_num=False, start=start, stop=stop, order=1, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 1)
        self.assertEqual(h_list[-1][0], 9)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        h_list = []
        for h in account.get_account_history(
            10, 10, use_block_num=False, order=-1, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 10)
        # self.assertEqual(h_list[-1][0], zero_element)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-11 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-1][1]["block"])
        h_list = []
        for h in account.get_account_history(
            10, 10, use_block_num=False, start=9, stop=1, order=-1, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(h_list[0][0], 9)
        self.assertEqual(h_list[-1][0], 1)
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-2 + zero_element][1]["block"])
        start = formatTimeString(h_list[0][1]["timestamp"])
        stop = formatTimeString(h_list[-1][1]["timestamp"])
        h_list = []
        for h in account.get_account_history(
            10, 10, start=start, stop=stop, order=-1, raw_output=True
        ):
            h_list.append(h)
        for i in range(len(h_list)):
            self.assertEqual(h_list[i][0], 9 - i)

        self.assertEqual(h_list[0][1]["block"], h_all_raw[-10 + zero_element][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-2 + zero_element][1]["block"])

    def test_history2(self):
        stm = self.bts
        account = Account("thecrazygm", steem_instance=stm)
        h_list = []
        max_index = account.virtual_op_count()
        for h in account.history(
            start=max_index - 4, stop=max_index, use_block_num=False, batch_size=2, raw_output=False
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i]["index"] - h_list[i - 1]["index"], 1)

        h_list = []
        for h in account.history(
            start=max_index - 4, stop=max_index, use_block_num=False, batch_size=6, raw_output=False
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i]["index"] - h_list[i - 1]["index"], 1)

        h_list = []
        for h in account.history(
            start=max_index - 4, stop=max_index, use_block_num=False, batch_size=2, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i][0] - h_list[i - 1][0], 1)

        h_list = []
        for h in account.history(
            start=max_index - 4, stop=max_index, use_block_num=False, batch_size=6, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i][0] - h_list[i - 1][0], 1)

    def test_history_index(self):
        stm = self.bts
        account = Account("thecrazygm", steem_instance=stm)
        h_list = []
        for h in account.history(
            start=1, stop=10, use_block_num=False, batch_size=10, raw_output=True
        ):
            h_list.append(h)
        for i in range(len(h_list)):
            self.assertEqual(h_list[i][0], i + 1)

        h_list = []
        for h in account.history(
            start=1, stop=10, use_block_num=False, batch_size=2, raw_output=True
        ):
            h_list.append(h)
        for i in range(len(h_list)):
            self.assertEqual(h_list[i][0], i + 1)

    def test_history_reverse2(self):
        stm = self.bts
        account = Account("thecrazygm", steem_instance=stm)
        h_list = []
        max_index = account.virtual_op_count()
        for h in account.history_reverse(
            start=max_index, stop=max_index - 4, use_block_num=False, batch_size=2, raw_output=False
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i]["index"] - h_list[i - 1]["index"], -1)

        h_list = []
        for h in account.history_reverse(
            start=max_index, stop=max_index - 4, use_block_num=False, batch_size=6, raw_output=False
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i]["index"] - h_list[i - 1]["index"], -1)

        h_list = []
        for h in account.history_reverse(
            start=max_index, stop=max_index - 4, use_block_num=False, batch_size=6, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i][0] - h_list[i - 1][0], -1)

        h_list = []
        for h in account.history_reverse(
            start=max_index, stop=max_index - 4, use_block_num=False, batch_size=2, raw_output=True
        ):
            h_list.append(h)
        self.assertEqual(len(h_list), 5)
        for i in range(1, 5):
            self.assertEqual(h_list[i][0] - h_list[i - 1][0], -1)

    def test_history_block_num(self):
        stm = self.bts
        zero_element = 0
        account = Account("thecrazygm", steem_instance=stm)
        h_all_raw = []
        for h in account.history_reverse(use_block_num=False, stop=-15, raw_output=True):
            h_all_raw.append(h)
        h_list = []
        self.assertTrue(len(h_all_raw) > 0)
        self.assertTrue(len(h_all_raw[0]) > 1)
        self.assertTrue("block" in h_all_raw[0][1])
        for h in account.history(
            start=h_all_raw[-1][1]["block"],
            stop=h_all_raw[0][1]["block"],
            use_block_num=True,
            batch_size=10,
            raw_output=True,
        ):
            h_list.append(h)
        # self.assertEqual(h_list[0][0], zero_element)
        self.assertEqual(h_list[-1][0], h_all_raw[0][0])
        self.assertEqual(h_list[0][1]["block"], h_all_raw[-1][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[0][1]["block"])
        h_list = []
        for h in account.history_reverse(
            start=h_all_raw[0][1]["block"],
            stop=h_all_raw[-1][1]["block"],
            use_block_num=True,
            batch_size=10,
            raw_output=True,
        ):
            h_list.append(h)
        # self.assertEqual(h_list[0][0], 10)

        self.assertEqual(h_list[0][0], h_all_raw[0][0])
        self.assertEqual(h_list[0][1]["block"], h_all_raw[0][1]["block"])
        self.assertEqual(h_list[-1][1]["block"], h_all_raw[-1][1]["block"])

    def test_account_props(self):
        account = self.account
        rep = account.get_reputation()
        self.assertTrue(isinstance(rep, float))
        vp = account.get_voting_power()
        self.assertTrue(vp >= 0)
        self.assertTrue(vp <= 100)
        sp = account.get_token_power()
        self.assertTrue(sp >= 0)
        vv = account.get_voting_value_SBD()
        self.assertTrue(vv >= 0)
        bw = account.get_bandwidth()
        # self.assertTrue(bw['used'] <= bw['allocated'])
        followers = account.get_followers()
        self.assertTrue(isinstance(followers, list))
        following = account.get_following()
        self.assertTrue(isinstance(following, list))
        count = account.get_follow_count()
        self.assertEqual(count["follower_count"], len(followers))
        self.assertEqual(count["following_count"], len(following))

    def test_MissingKeyError(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.convert("1 SBD")
        with self.assertRaises(exceptions.MissingKeyError):
            tx.sign()

    def test_withdraw_vesting(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.withdraw_vesting("100 VESTS")
        self.assertEqual((tx["operations"][0][0]), "withdraw_vesting")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["account"])

    def test_delegate_vesting_shares(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.delegate_vesting_shares("test1", "100 VESTS")
        self.assertEqual((tx["operations"][0][0]), "delegate_vesting_shares")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["delegator"])

    def test_claim_reward_balance(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.claim_reward_balance()
        self.assertEqual((tx["operations"][0][0]), "claim_reward_balance")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["account"])

    def test_cancel_transfer_from_savings(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.cancel_transfer_from_savings(0)
        self.assertEqual((tx["operations"][0][0]), "cancel_transfer_from_savings")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])

    def test_transfer_from_savings(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.transfer_from_savings(1, "STEEM", "")
        self.assertEqual((tx["operations"][0][0]), "transfer_from_savings")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])

    def test_transfer_to_savings(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.transfer_to_savings(1, "STEEM", "")
        self.assertEqual((tx["operations"][0][0]), "transfer_to_savings")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])

    def test_convert(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.convert("1 SBD")
        self.assertEqual((tx["operations"][0][0]), "convert")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["owner"])

    def test_proxy(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.setproxy(proxy="gtg")
        self.assertEqual((tx["operations"][0][0]), "account_witness_proxy")
        op = tx["operations"][0][1]
        self.assertIn("gtg", op["proxy"])

    def test_transfer_to_vesting(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.transfer_to_vesting("1 STEEM")
        self.assertEqual((tx["operations"][0][0]), "transfer_to_vesting")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])

        w.blockchain.txbuffer.clear()
        tx = w.transfer_to_vesting("1 STEEM", skip_account_check=True)
        self.assertEqual((tx["operations"][0][0]), "transfer_to_vesting")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])

    def test_transfer(self):
        w = self.account
        w.blockchain.txbuffer.clear()
        tx = w.transfer("thecrazygm", "1", "STEEM")
        self.assertEqual((tx["operations"][0][0]), "transfer")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])
        self.assertIn("thecrazygm", op["to"])

        w.blockchain.txbuffer.clear()
        tx = w.transfer("thecrazygm", "1", "STEEM", skip_account_check=True)
        self.assertEqual((tx["operations"][0][0]), "transfer")
        op = tx["operations"][0][1]
        self.assertIn("thecrazygm", op["from"])
        self.assertIn("thecrazygm", op["to"])

    def test_json_export(self):
        account = Account("thecrazygm", steem_instance=self.bts)
        if account.blockchain.rpc.get_use_appbase():
            content = self.bts.rpc.find_accounts({"accounts": [account["name"]]}, api="database")[
                "accounts"
            ][0]
        else:
            content = self.bts.rpc.get_accounts([account["name"]])[0]
        keys = list(content.keys())
        json_content = account.json()
        exclude_list = [
            "owner_challenged",
            "average_bandwidth",
        ]  # ['json_metadata', 'reputation', 'active_votes', 'savings_sbd_seconds']
        for k in keys:
            if k not in exclude_list:
                if isinstance(content[k], dict) and isinstance(json_content[k], list):
                    content_list = [
                        content[k]["amount"],
                        content[k]["precision"],
                        content[k]["nai"],
                    ]
                    self.assertEqual(content_list, json_content[k])
                else:
                    self.assertEqual(content[k], json_content[k])

    def test_estimate_virtual_op_num(self):
        stm = self.bts
        account = Account("gtg", steem_instance=stm)
        block_num = 21248120
        block = Block(block_num, steem_instance=stm)
        op_num1 = account.estimate_virtual_op_num(block.time(), stop_diff=1, max_count=100)
        op_num2 = account.estimate_virtual_op_num(block_num, stop_diff=1, max_count=100)
        op_num3 = account.estimate_virtual_op_num(block_num, stop_diff=100, max_count=100)
        op_num4 = account.estimate_virtual_op_num(block_num, stop_diff=0.00001, max_count=100)
        self.assertTrue(abs(op_num1 - op_num2) < 2)
        self.assertTrue(abs(op_num1 - op_num4) < 2)
        self.assertTrue(abs(op_num1 - op_num3) < 200)
        block_diff1 = 0
        block_diff2 = 0
        for h in account.get_account_history(op_num4 - 1, 0):
            block_diff1 = block_num - h["block"]
        for h in account.get_account_history(op_num4 + 1, 0):
            block_diff2 = block_num - h["block"]
        self.assertTrue(block_diff1 > 0)
        self.assertTrue(block_diff2 <= 0)

    def test_estimate_virtual_op_num2(self):
        account = self.account
        h_all_raw = []
        for h in account.history(raw_output=False):
            h_all_raw.append(h)
        last_block = h_all_raw[0]["block"]
        i = 1
        for op in h_all_raw[1:5]:
            new_block = op["block"]
            block_num = last_block + int((new_block - last_block) / 2)
            op_num = account.estimate_virtual_op_num(block_num, stop_diff=0.1, max_count=100)
            if op_num > 0:
                op_num -= 1
            self.assertTrue(op_num <= i)
            i += 1
            last_block = new_block

    def test_comment_history(self):
        account = self.account
        comments = []
        for c in account.comment_history(limit=1):
            comments.append(c)
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0]["author"], account["name"])
        self.assertTrue(comments[0].is_comment())
        self.assertTrue(comments[0].depth > 0)

    def test_blog_history(self):
        account = Account("thecrazygm", steem_instance=self.bts)
        posts = []
        for p in account.blog_history(limit=5):
            if p["author"] != account["name"]:
                continue
            posts.append(p)
        self.assertTrue(len(posts) >= 1)
        self.assertEqual(posts[0]["author"], account["name"])
        self.assertTrue(posts[0].is_main_post())
        self.assertTrue(posts[0].depth == 0)

    def test_reply_history(self):
        account = self.account
        replies = []
        for r in account.reply_history(limit=1):
            replies.append(r)
        # self.assertEqual(len(replies), 1)
        if len(replies) > 0:
            self.assertTrue(replies[0].is_comment())
            self.assertTrue(replies[0].depth > 0)

    def test_get_vote_pct_for_vote_value(self):
        account = self.account
        for vote_pwr in range(5, 100, 5):
            self.assertTrue(
                9900
                <= account.get_vote_pct_for_vote_value(
                    account.get_voting_value(voting_power=vote_pwr), voting_power=vote_pwr
                )
                <= 11000
            )

    def test_extract_account_name(self):
        stm = self.bts
        account = Account("thecrazygm", steem_instance=stm)
        self.assertEqual(extract_account_name(account), "thecrazygm")
        self.assertEqual(extract_account_name("thecrazygm"), "thecrazygm")
        self.assertEqual(extract_account_name({"name": "thecrazygm"}), "thecrazygm")
        self.assertEqual(extract_account_name(""), "")



# -*- coding: utf-8 -*-
import logging

from nectarstorage import SqliteConfigurationStore, SqliteEncryptedKeyStore

from .nodelist import NodeList

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.StreamHandler())


timeformat = "%Y%m%d-%H%M%S"


def generate_config_store(config, blockchain="hive"):
    #: Default configuration
    nodelist = NodeList()
    if blockchain == "hive":
        nodes = nodelist.get_hive_nodes(testnet=False)
    elif blockchain == "steem":
        nodes = nodelist.get_steem_nodes(testnet=False)
    else:
        nodes = []

    config.setdefault("node", nodes)
    config.setdefault("default_chain", blockchain)
    config.setdefault("password_storage", "environment")
    config.setdefault("rpcpassword", "")
    config.setdefault("rpcuser", "")
    config.setdefault("order-expiration", 7 * 24 * 60 * 60)
    config.setdefault("client_id", "")
    config.setdefault("sc2_client_id", None)
    config.setdefault("hs_client_id", None)
    config.setdefault("hot_sign_redirect_uri", None)
    config.setdefault("sc2_api_url", "https://api.steemconnect.com/api/")
    config.setdefault("oauth_base_url", "https://api.steemconnect.com/oauth2/")
    config.setdefault("hs_api_url", "https://hivesigner.com/api/")
    config.setdefault("hs_oauth_base_url", "https://hivesigner.com/oauth2/")
    config.setdefault("default_canonical_url", "https://hive.blog")
    config.setdefault("default_path", "48'/13'/0'/0'/0'")
    config.setdefault("use_condenser", True)
    config.setdefault("use_tor", False)
    return config


def get_default_config_store(*args, **kwargs):
    return generate_config_store(SqliteConfigurationStore, blockchain="hive")(*args, **kwargs)


def get_default_key_store(config, *args, **kwargs):
    return SqliteEncryptedKeyStore(config=config, **kwargs)



# -*- coding: utf-8 -*-
import json
import logging
from datetime import date, datetime, time
from typing import Union

from prettytable import PrettyTable

from nectar.instance import shared_blockchain_instance

from .blockchainobject import BlockchainObject
from .exceptions import AccountDoesNotExistsException, OfflineHasNoRPCException
from .utils import (
    addTzInfo,
    formatTimeString,
)

log = logging.getLogger(__name__)


class Community(BlockchainObject):
    """A class representing a Hive community with methods to interact with it.

    This class provides an interface to access and manipulate community data on the Hive blockchain.
    It extends BlockchainObject and provides additional community-specific functionality.

    Args:
        community: Either a community name (str) or a dictionary containing community data
        observer: Observer account for personalized results (default: "")
        full: If True, fetch full community data (default: True)
        lazy: If True, use lazy loading (default: False)
        blockchain_instance: Hive or Steem instance for blockchain access
        **kwargs: Additional arguments including 'hive_instance' or 'steem_instance'

    Attributes:
        type_id (int): Type identifier for blockchain objects (2 for communities)

    Example:
        >>> from nectar.community import Community
        >>> from nectar import Hive
        >>> from nectar.nodelist import NodeList
        >>> nodelist = NodeList()
        >>> nodelist.update_nodes()
        >>> stm = Hive(node=nodelist.get_hive_nodes())
        >>> community = Community("hive-139531", blockchain_instance=stm)
        >>> print(community)
        <Community hive-139531>

    Note:
        This class includes caching to reduce API server load. Use refresh() to update
        the data and clear_cache() to clear the cache.
    """

    type_id = 2

    def __init__(
        self,
        community: Union[str, dict],
        observer: str = "",
        full: bool = True,
        lazy: bool = False,
        blockchain_instance=None,
        **kwargs,
    ) -> None:
        """Initialize a Community instance.

        Args:
            community: Either a community name (str) or a dictionary containing community data
            observer: Observer account for personalized results (default: "")
            full: If True, fetch full community data (default: True)
            lazy: If True, use lazy loading (default: False)
            blockchain_instance: Hive or Steem instance for blockchain access
            **kwargs: Additional arguments including 'hive_instance' or 'steem_instance'
        """
        self.full = full
        self.lazy = lazy
        self.observer = observer
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(community, dict):
            community = self._parse_json_data(community)
        super(Community, self).__init__(
            community, lazy=lazy, full=full, id_item="name", blockchain_instance=blockchain_instance
        )

    def refresh(self) -> None:
        """Refresh the community's data from the blockchain.

        This method updates the community's data by fetching the latest information
        from the blockchain. It raises an exception if the community doesn't exist.

        Raises:
            AccountDoesNotExistsException: If the community doesn't exist on the blockchain
            OfflineHasNoRPCException: If not connected to the blockchain
        """
        if not self.blockchain.is_connected():
            return
        self.blockchain.rpc.set_next_node_on_empty_reply(True)
        community = self.blockchain.rpc.get_community(
            {"name": self.identifier, "observer": self.observer}, api="bridge"
        )

        if not community:
            raise AccountDoesNotExistsException(self.identifier)
        community = self._parse_json_data(community)
        self.identifier = community["name"]
        # self.blockchain.refresh_data()

        super(Community, self).__init__(
            community,
            id_item="name",
            lazy=self.lazy,
            full=self.full,
            blockchain_instance=self.blockchain,
        )

    def _parse_json_data(self, community: dict) -> dict:
        """Parse and convert community JSON data into proper Python types.

        This internal method converts string representations of numbers to integers
        and parses date strings into datetime objects with timezone information.

        Args:
            community: Dictionary containing raw community data from the API

        Returns:
            dict: Processed community data with proper Python types
        """
        # Convert string numbers to integers
        int_fields = [
            "sum_pending",
            "subscribers",
            "num_pending",
            "num_authors",
        ]
        for field in int_fields:
            if field in community and isinstance(community.get(field), str):
                community[field] = int(community.get(field, 0))

        # Parse date strings into datetime objects
        date_fields = ["created_at"]
        for field in date_fields:
            if field in community and isinstance(community.get(field), str):
                community[field] = addTzInfo(
                    datetime.strptime(
                        community.get(field, "1970-01-01 00:00:00"), "%Y-%m-%d %H:%M:%S"
                    )
                )

        return community

    def json(self) -> dict:
        """Convert the community data to a JSON-serializable dictionary.

        This method prepares the community data for JSON serialization by converting
        non-JSON-serializable types (like datetime objects) to strings.

        Returns:
            dict: A dictionary containing the community data in a JSON-serializable format
        """
        output = self.copy()

        # Convert integer fields to strings for JSON serialization
        int_fields = [
            "sum_pending",
            "subscribers",
            "num_pending",
            "num_authors",
        ]

        # Fields that should only be converted if non-zero
        int_non_zero_fields = []

        # Convert regular integer fields
        for field in int_fields:
            if field in output and isinstance(output[field], int):
                output[field] = str(output[field])

        # Convert non-zero integer fields
        for field in int_non_zero_fields:
            if field in output and isinstance(output[field], int) and output[field] != 0:
                output[field] = str(output[field])

        # Convert datetime fields to ISO format strings
        date_fields = ["created_at"]
        for field in date_fields:
            if field in output:
                date_val = output.get(field, datetime(1970, 1, 1, 0, 0))
                if isinstance(date_val, (datetime, date, time)):
                    output[field] = formatTimeString(date_val).replace("T", " ")
                else:
                    output[field] = date_val
        return json.loads(str(json.dumps(output)))

    def get_community_roles(self, limit: int = 100, last: str = None) -> list:
        """Lists community roles

        Args:
            limit: Maximum number of roles to return (default: 100)
            last: Account name of the last role from previous page for pagination

        Returns:
            list: List of community roles

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
        """
        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        params = {"community": community, "limit": limit}
        if last is not None:
            params["last"] = last

        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        return self.blockchain.rpc.list_community_roles(params, api="bridge")

    def get_subscribers(self, limit: int = 100, last: str = None) -> list:
        """Returns subscribers

        Args:
            limit: Maximum number of subscribers to return (default: 100)
            last: Account name of the last subscriber from previous page for pagination

        Returns:
            list: List of subscribers

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
        """
        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        params = {"community": community, "limit": limit}
        if last is not None:
            params["last"] = last

        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        return self.blockchain.rpc.list_subscribers(params, api="bridge")

    def get_activities(self, limit: int = 100, last_id: str = None) -> list:
        """Returns community activity

        Args:
            limit: Maximum number of activities to return (default: 100)
            last_id: ID of the last activity from previous page for pagination

        Returns:
            list: List of community activities

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
        """
        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        params = {"account": community, "limit": limit}
        if last_id is not None:
            params["last_id"] = last_id

        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        return self.blockchain.rpc.account_notifications(params, api="bridge")

    def get_ranked_posts(
        self,
        observer: str = None,
        limit: int = 100,
        start_author: str = None,
        start_permlink: str = None,
        sort: str = "created",
    ) -> list:
        """Returns community posts

        Args:
            observer: Account name of the observer (optional)
            limit: Maximum number of posts to return (default: 100)
            start_author: Author of the post to start from for pagination (optional)
            start_permlink: Permlink of the post to start from for pagination (optional)
            sort: Sort order (default: "created")

        Returns:
            list: List of community posts

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
        """
        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        params = {"tag": community, "limit": limit, "sort": sort}

        if observer is not None:
            params["observer"] = observer
        if start_author is not None:
            params["start_author"] = start_author
        if start_permlink is not None:
            params["start_permlink"] = start_permlink

        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        return self.blockchain.rpc.get_ranked_posts(params, api="bridge")

    def set_role(self, account: str, role: str, mod_account: str) -> dict:
        """Set role for a given account in the community.

        Args:
            account: Account name to set the role for
            role: Role to assign (member, mod, admin, owner, or guest)
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If role is not one of the allowed values
        """
        valid_roles = {"member", "mod", "admin", "owner", "guest"}
        if role.lower() not in valid_roles:
            raise ValueError(f"Invalid role. Must be one of: {', '.join(valid_roles)}")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "setRole",
            {
                "community": community,
                "account": account,
                "role": role.lower(),
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def set_user_title(self, account: str, title: str, mod_account: str) -> dict:
        """Set the title for a given account in the community.

        Args:
            account: Account name to set the title for
            title: Title to assign to the account
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If account or title is empty
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")

        if not title or not isinstance(title, str):
            raise ValueError("Title must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "setUserTitle",
            {
                "community": community,
                "account": account,
                "title": title.strip(),
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def mute_post(self, account: str, permlink: str, notes: str, mod_account: str) -> dict:
        """Mutes a post in the community.

        Args:
            account: Author of the post to mute
            permlink: Permlink of the post to mute
            notes: Reason for muting the post
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")
        if not permlink or not isinstance(permlink, str):
            raise ValueError("Permlink must be a non-empty string")
        if not isinstance(notes, str):
            raise ValueError("Notes must be a string")
        if not mod_account or not isinstance(mod_account, str):
            raise ValueError("Moderator account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "mutePost",
            {
                "community": community,
                "account": account,
                "permlink": permlink,
                "notes": notes.strip(),
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def unmute_post(self, account: str, permlink: str, notes: str, mod_account: str) -> dict:
        """Unmute a previously muted post in the community.

        Args:
            account: Author of the post to unmute
            permlink: Permlink of the post to unmute
            notes: Reason for unmuting the post
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")
        if not permlink or not isinstance(permlink, str):
            raise ValueError("Permlink must be a non-empty string")
        if not isinstance(notes, str):
            raise ValueError("Notes must be a string")
        if not mod_account or not isinstance(mod_account, str):
            raise ValueError("Moderator account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "unmutePost",
            {
                "community": community,
                "account": account,
                "permlink": permlink,
                "notes": notes.strip(),
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def update_props(
        self,
        title: str,
        about: str,
        is_nsfw: bool,
        description: str,
        flag_text: str,
        admin_account: str,
    ) -> dict:
        """Update community properties.

        Args:
            title: New title for the community (must be non-empty)
            about: Brief description of the community
            is_nsfw: Whether the community contains NSFW content
            description: Detailed description of the community
            flag_text: Text shown when flagging content in this community
            admin_account: Account name of the admin performing this action

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not title or not isinstance(title, str):
            raise ValueError("Title must be a non-empty string")
        if not isinstance(about, str):
            about = ""
        if not isinstance(description, str):
            description = ""
        if not isinstance(flag_text, str):
            flag_text = ""
        if not admin_account or not isinstance(admin_account, str):
            raise ValueError("Admin account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "updateProps",
            {
                "community": community,
                "props": {
                    "title": title.strip(),
                    "about": about.strip(),
                    "is_nsfw": bool(is_nsfw),
                    "description": description.strip(),
                    "flag_text": flag_text.strip(),
                },
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[admin_account]
        )

    def subscribe(self, account: str) -> dict:
        """Subscribe an account to this community.

        The account that calls this method will be subscribed to the community.
        The same account must be used to sign the transaction.

        Args:
            account: Account name that wants to subscribe to the community

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If account is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "subscribe",
            {
                "community": community,
            },
        ]
        return self.blockchain.custom_json("community", json_body, required_posting_auths=[account])

    def pin_post(self, account: str, permlink: str, mod_account: str) -> dict:
        """Pin a post to the top of the community feed.

        This method allows community moderators to pin a specific post to the top of the
        community's feed. The post will remain pinned until it is manually unpinned.

        Args:
            account: Author of the post to pin
            permlink: Permlink of the post to pin
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")
        if not permlink or not isinstance(permlink, str):
            raise ValueError("Permlink must be a non-empty string")
        if not mod_account or not isinstance(mod_account, str):
            raise ValueError("Moderator account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "pinPost",
            {
                "community": community,
                "account": account,
                "permlink": permlink,
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def unsubscribe(self, account: str) -> dict:
        """Unsubscribe an account from this community.

        The account that calls this method will be unsubscribed from the community.
        The same account must be used to sign the transaction.

        Args:
            account: Account name that wants to unsubscribe from the community

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If account is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "unsubscribe",
            {
                "community": community,
            },
        ]
        return self.blockchain.custom_json("community", json_body, required_posting_auths=[account])

    def unpin_post(self, account: str, permlink: str, mod_account: str) -> dict:
        """Remove a post from being pinned at the top of the community feed.

        This method allows community moderators to unpin a previously pinned post.
        After unpinning, the post will return to its normal position in the feed.

        Args:
            account: Author of the post to unpin
            permlink: Permlink of the post to unpin
            mod_account: Account name of the moderator performing this action (must be mod or higher)

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")
        if not permlink or not isinstance(permlink, str):
            raise ValueError("Permlink must be a non-empty string")
        if not mod_account or not isinstance(mod_account, str):
            raise ValueError("Moderator account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "unpinPost",
            {
                "community": community,
                "account": account,
                "permlink": permlink,
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[mod_account]
        )

    def flag_post(self, account: str, permlink: str, notes: str, reporter: str) -> dict:
        """Report a post to the community moderators for review.

        This method allows community members to flag posts that may violate
        community guidelines. The post will be added to the community's
        review queue for moderators to evaluate.

        Args:
            account: Author of the post being reported
            permlink: Permlink of the post being reported
            notes: Explanation of why the post is being reported
            reporter: Account name of the user reporting the post

        Returns:
            dict: Transaction result

        Raises:
            OfflineHasNoRPCException: If not connected to the blockchain
            ValueError: If any required parameter is invalid
        """
        if not account or not isinstance(account, str):
            raise ValueError("Account must be a non-empty string")
        if not permlink or not isinstance(permlink, str):
            raise ValueError("Permlink must be a non-empty string")
        if not notes or not isinstance(notes, str):
            raise ValueError("Notes must be a string")
        if not reporter or not isinstance(reporter, str):
            raise ValueError("Reporter account must be a non-empty string")

        community = self["name"]
        if not self.blockchain.is_connected():
            raise OfflineHasNoRPCException("No RPC available in offline mode!")

        json_body = [
            "flagPost",
            {
                "community": community,
                "account": account,
                "permlink": permlink,
                "notes": notes.strip(),
            },
        ]
        return self.blockchain.custom_json(
            "community", json_body, required_posting_auths=[reporter]
        )


class CommunityObject(list):
    """A list-like container for Community objects with additional utility methods."""

    def printAsTable(self) -> None:
        """Print a formatted table of communities with key metrics.

        The table includes the following columns:
        - Nr.: Sequential number
        - Name: Community name
        - Title: Community title
        - lang: Language code
        - subscribers: Number of subscribers
        - sum_pending: Sum of pending payouts
        - num_pending: Number of pending posts
        - num_authors: Number of unique authors
        """
        t = PrettyTable(
            [
                "Nr.",
                "Name",
                "Title",
                "lang",
                "subscribers",
                "sum_pending",
                "num_pending",
                "num_authors",
            ]
        )
        t.align = "l"
        count = 0
        for community in self:
            count += 1
            t.add_row(
                [
                    str(count),
                    community["name"],
                    community["title"],
                    community["lang"],
                    community["subscribers"],
                    community["sum_pending"],
                    community["num_pending"],
                    community["num_authors"],
                ]
            )
        print(t)


class Communities(CommunityObject):
    """A list of communities with additional querying capabilities.

    This class extends CommunityObject to provide methods for fetching and
    searching communities from the blockchain.

    Args:
        sort: Sort order for communities (default: "rank")
        observer: Observer account for personalized results (optional)
        last: Last community name for pagination (optional)
        limit: Maximum number of communities to fetch (default: 100)
        lazy: If True, use lazy loading (default: False)
        full: If True, fetch full community data (default: True)
        blockchain_instance: Hive or Steem instance to use for blockchain access
        **kwargs: Additional arguments including 'steem_instance' or 'hive_instance'
    """

    def __init__(
        self,
        sort: str = "rank",
        observer: str = None,
        last: str = None,
        limit: int = 100,
        lazy: bool = False,
        full: bool = True,
        blockchain_instance=None,
        **kwargs,
    ) -> None:
        """Initialize the Communities list with the given parameters.

        Args:
            sort: Sort order for communities (default: "rank")
            observer: Observer account for personalized results (optional)
            last: Last community name for pagination (optional)
            limit: Maximum number of communities to fetch (default: 100)
            lazy: If True, use lazy loading (default: False)
            full: If True, fetch full community data (default: True)
            blockchain_instance: Hive or Steem instance to use for blockchain access
            **kwargs: Additional arguments including 'steem_instance' or 'hive_instance'
        """
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if not self.blockchain.is_connected():
            return

        communities = []
        community_cnt = 0
        batch_limit = min(100, limit)  # Ensure we don't exceed the limit

        while community_cnt < limit:
            self.blockchain.rpc.set_next_node_on_empty_reply(False)
            batch = self.blockchain.rpc.list_communities(
                {"sort": sort, "observer": observer, "last": last, "limit": batch_limit},
                api="bridge",
            )
            if not batch:  # No more communities to fetch
                break

            communities.extend(batch)
            community_cnt += len(batch)
            last = communities[-1]["name"]

            # Adjust batch size for the next iteration if needed
            if community_cnt + batch_limit > limit:
                batch_limit = limit - community_cnt

        super(Communities, self).__init__(
            [
                Community(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in communities[:limit]  # Ensure we don't exceed the limit
            ]
        )

    def search_title(self, title: str) -> "CommunityObject":
        """Search for communities with titles containing the given string.

        The search is case-insensitive.

        Args:
            title: Text to search for in community titles

        Returns:
            CommunityObject: A new CommunityObject containing matching communities
        """
        if not title or not isinstance(title, str):
            raise ValueError("Title must be a non-empty string")

        ret = CommunityObject()
        title_lower = title.lower()
        for community in self:
            if title_lower in community["title"].lower():
                ret.append(community)
        return ret



# -*- coding: utf-8 -*-
import base64
import hashlib
import json
import random
import struct
from binascii import hexlify
from datetime import datetime, timezone

import requests

from nectargraphenebase.ecdsasig import sign_message

from .account import Account
from .instance import shared_blockchain_instance

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin


class Conveyor(object):
    """Class to access Steemit Conveyor instances:
    https://github.com/steemit/conveyor

    Description from the official documentation:

    * Feature flags: "Feature flags allows our apps (condenser mainly) to
      hide certain features behind flags."
    * User data: "Conveyor is the central point for storing sensitive user
      data (email, phone, etc). No other services should store this data
      and should instead query for it here every time."
    * User tags: "Tagging mechanism for other services, allows defining and
      assigning tags to accounts (or other identifiers) and querying for
      them."

    Not contained in the documentation, but implemented and working:

    * Draft handling: saving, listing and removing post drafts
      consisting of a post title and a body.

    The underlying RPC authentication and request signing procedure is
    described here: https://github.com/steemit/rpc-auth

    """

    def __init__(self, url="https://conveyor.steemit.com", blockchain_instance=None, **kwargs):
        """Initialize a Conveyor instance
        :param str url: (optional) URL to the Conveyor API, defaults to
            https://conveyor.steemit.com
        :param nectar.steem.Steem blockchain_instance: Steem instance

        """

        self.url = url
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.steem = blockchain_instance or shared_blockchain_instance()
        self.id = 0
        self.ENCODING = "utf-8"
        self.TIMEFORMAT = "%Y-%m-%dT%H:%M:%S.%f"
        self.K = hashlib.sha256(bytes("steem_jsonrpc_auth", self.ENCODING)).digest()

    def prehash_message(self, timestamp, account, method, params, nonce):
        """Prepare a hash for the Conveyor API request with SHA256 according
        to https://github.com/steemit/rpc-auth
        Hashing of `second` is then done inside `ecdsasig.sign_message()`.

        :param str timestamp: valid iso8601 datetime ending in "Z"
        :param str account: valid steem blockchain account name
        :param str method: Conveyor method name to be called
        :param bytes param: base64 encoded request parameters
        :param bytes nonce: random 8 bytes

        """
        first = hashlib.sha256(bytes(timestamp + account + method + params, self.ENCODING))
        return self.K + first.digest() + nonce

    def _request(self, account, method, params, key):
        """Assemble the request, hash it, sign it and send it to the Conveyor
        instance. Returns the server response as JSON.

        :param str account: account name
        :param str method: Conveyor method name to be called
        :param dict params: request parameters as `dict`
        :param str key: Steem posting key for signing

        """
        params_bytes = bytes(json.dumps(params), self.ENCODING)
        params_enc = base64.b64encode(params_bytes).decode(self.ENCODING)
        timestamp = datetime.now(timezone.utc).strftime(self.TIMEFORMAT)[:-3] + "Z"
        nonce_int = random.getrandbits(64)
        nonce_bytes = struct.pack(">Q", nonce_int)  # 64bit ULL, big endian
        nonce_str = "%016x" % (nonce_int)

        message = self.prehash_message(timestamp, account, method, params_enc, nonce_bytes)
        signature = sign_message(message, key)
        signature_hex = hexlify(signature).decode(self.ENCODING)

        request = {
            "jsonrpc": "2.0",
            "id": self.id,
            "method": method,
            "params": {
                "__signed": {
                    "account": account,
                    "nonce": nonce_str,
                    "params": params_enc,
                    "signatures": [signature_hex],
                    "timestamp": timestamp,
                }
            },
        }
        r = requests.post(self.url, data=json.dumps(request))
        self.id += 1
        return r.json()

    def _conveyor_method(self, account, signing_account, method, params):
        """Wrapper function to handle account and key lookups

        :param str account: name of the addressed account
        :param str signing_account: name of the account to sign the request
        :param method: Conveyor method name to be called
        :params dict params: request parameters as `dict`

        """
        account = Account(account, blockchain_instance=self.steem)
        if signing_account is None:
            signer = account
        else:
            signer = Account(signing_account, blockchain_instance=self.steem)
        if "posting" not in signer:
            signer.refresh()
        if "posting" not in signer:
            raise AssertionError("Could not access posting permission")
        for authority in signer["posting"]["key_auths"]:
            posting_wif = self.steem.wallet.getPrivateKeyForPublicKey(authority[0])
        return self._request(account["name"], method, params, posting_wif)

    def get_user_data(self, account, signing_account=None):
        """Get the account's email address and phone number. The request has to be
        signed by the requested account or an admin account.

        :param str account: requested account
        :param str signing_account: (optional) account to sign the
            request. If unset, `account` is used.

        Example:

        .. code-block:: python

            from nectar import Steem
            from nectar.conveyor import Conveyor
            s = Steem(keys=["5JPOSTINGKEY"])
            c = Conveyor(blockchain_instance=s)
            print(c.get_user_data('accountname'))

        """
        account = Account(account, blockchain_instance=self.steem)
        user_data = self._conveyor_method(
            account, signing_account, "conveyor.get_user_data", [account["name"]]
        )
        if "result" in user_data:
            return user_data["result"]
        else:
            return user_data

    def set_user_data(self, account, params, signing_account=None):
        """Set the account's email address and phone number. The request has to be
        signed by an admin account.

        :param str account: requested account
        :param dict param: user data to be set
        :param str signing_account: (optional) account to sign the
            request. If unset, `account` is used.

        Example:

        .. code-block:: python

            from nectar import Steem
            from nectar.conveyor import Conveyor
            s = Steem(keys=["5JADMINPOSTINGKEY"])
            c = Conveyor(blockchain_instance=s)
            userdata = {'email': 'foo@bar.com', 'phone':'+123456789'}
            c.set_user_data('accountname', userdata, 'adminaccountname')

        """
        return self._conveyor_method(account, signing_account, "conveyor.set_user_data", [params])

    def get_feature_flags(self, account, signing_account=None):
        """Get the account's feature flags. The request has to be signed by the
        requested account or an admin account.

        :param str account: requested account
        :param str signing_account: (optional) account to sign the
            request. If unset, `account` is used.

        Example:

        .. code-block:: python

            from nectar import Steem
            from nectar.conveyor import Conveyor
            s = Steem(keys=["5JPOSTINGKEY"])
            c = Conveyor(blockchain_instance=s)
            print(c.get_feature_flags('accountname'))

        """
        account = Account(account, blockchain_instance=self.steem)
        feature_flags = self._conveyor_method(
            account, signing_account, "conveyor.get_feature_flags", [account["name"]]
        )
        if "result" in feature_flags:
            return feature_flags["result"]
        else:
            return feature_flags

    def get_feature_flag(self, account, flag, signing_account=None):
        """Test if a specific feature flag is set for an account. The request
        has to be signed by the requested account or an admin account.

        :param str account: requested account
        :param str flag: flag to be tested
        :param str signing_account: (optional) account to sign the
            request. If unset, `account` is used.

        Example:

        .. code-block:: python

            from nectar import Steem
            from nectar.conveyor import Conveyor
            s = Steem(keys=["5JPOSTINGKEY"])
            c = Conveyor(blockchain_instance=s)
            print(c.get_feature_flag('accountname', 'accepted_tos'))

        """
        account = Account(account, blockchain_instance=self.steem)
        return self._conveyor_method(
            account, signing_account, "conveyor.get_feature_flag", [account["name"], flag]
        )

    def save_draft(self, account, title, body):
        """Save a draft in the Conveyor database

        :param str account: requested account
        :param str title: draft post title
        :param str body: draft post body

        """
        account = Account(account, blockchain_instance=self.steem)
        draft = {"title": title, "body": body}
        return self._conveyor_method(account, None, "conveyor.save_draft", [account["name"], draft])

    def list_drafts(self, account):
        """List all saved drafts from `account`

        :param str account: requested account

        Sample output:

        .. code-block:: js

            {
                'jsonrpc': '2.0', 'id': 2, 'result': [
                    {'title': 'draft-title', 'body': 'draft-body',
                     'uuid': '06497e1e-ac30-48cb-a069-27e1672924c9'}
                ]
            }

        """
        account = Account(account, blockchain_instance=self.steem)
        return self._conveyor_method(account, None, "conveyor.list_drafts", [account["name"]])

    def remove_draft(self, account, uuid):
        """Remove a draft from the Conveyor database

        :param str account: requested account
        :param str uuid: draft identifier as returned from
            `list_drafts`

        """
        account = Account(account, blockchain_instance=self.steem)
        return self._conveyor_method(
            account, None, "conveyor.remove_draft", [account["name"], uuid]
        )

    def healthcheck(self):
        """Get the Conveyor status

        Sample output:

        .. code-block:: js

            {
                'ok': True, 'version': '1.1.1-4d28e36-1528725174',
                'date': '2018-07-21T12:12:25.502Z'
            }

        """
        url = urljoin(self.url, "/.well-known/healthcheck.json")
        r = requests.get(url)
        return r.json()



# -*- coding: utf-8 -*-
import json
import logging
import re
import ssl

from nectargraphenebase.chains import known_chains
from nectargraphenebase.version import version as nectar_version

from .exceptions import (
    CallRetriesReached,
    RPCConnection,
    RPCError,
    RPCErrorDoRetry,
    UnauthorizedError,
    WorkingNodeMissing,
)
from .node import Nodes
from .rpcutils import get_api_name, get_query, is_network_appbase_ready

WEBSOCKET_MODULE = None
if not WEBSOCKET_MODULE:
    try:
        import websocket
        from websocket._exceptions import (
            WebSocketConnectionClosedException,
            WebSocketTimeoutException,
        )

        WEBSOCKET_MODULE = "websocket"
    except ImportError:
        WEBSOCKET_MODULE = None
REQUEST_MODULE = None
if not REQUEST_MODULE:
    try:
        import requests
        from requests.adapters import HTTPAdapter
        from requests.exceptions import ConnectionError
        from requests.packages.urllib3.util.retry import Retry

        REQUEST_MODULE = "requests"
    except ImportError:
        REQUEST_MODULE = None

log = logging.getLogger(__name__)


class SessionInstance(object):
    """Singleton for the Session Instance"""

    instance = None


def set_session_instance(instance):
    """Set session instance"""
    SessionInstance.instance = instance


def shared_session_instance():
    """Get session instance"""
    if REQUEST_MODULE is None:
        raise Exception("Requests module is not available.")
    if not SessionInstance.instance:
        SessionInstance.instance = requests.Session()
    return SessionInstance.instance


def create_ws_instance(use_ssl=True, enable_multithread=True):
    """Get websocket instance"""
    if WEBSOCKET_MODULE is None:
        raise Exception("WebSocket module is not available.")
    if use_ssl:
        ssl_defaults = ssl.get_default_verify_paths()
        sslopt_ca_certs = {"ca_certs": ssl_defaults.cafile}
        return websocket.WebSocket(sslopt=sslopt_ca_certs, enable_multithread=enable_multithread)
    else:
        return websocket.WebSocket(enable_multithread=enable_multithread)


class GrapheneRPC(object):
    """
    This class allows calling API methods synchronously, without callbacks.

    It logs warnings and errors.

    :param str urls: Either a single Websocket/Http URL, or a list of URLs
    :param str user: Username for Authentication
    :param str password: Password for Authentication
    :param int num_retries: Number of retries for node connection (default is 100)
    :param int num_retries_call: Number of retries for RPC calls on node error (default is 5)
    :param int timeout: Timeout setting for HTTP nodes (default is 60)
    :param bool autoconnect: Automatically connect on initialization (default is True)
    :param bool use_condenser: Use the old condenser_api RPC protocol
    :param bool use_tor: Use Tor proxy for connections
    :param dict custom_chains: Custom chains to add to known chains
    """

    def __init__(self, urls, user=None, password=None, **kwargs):
        """Initialize the RPC client."""
        self.rpc_methods = {"offline": -1, "ws": 0, "jsonrpc": 1, "wsappbase": 2, "appbase": 3}
        self.current_rpc = self.rpc_methods["ws"]
        self._request_id = 0
        self.timeout = kwargs.get("timeout", 60)
        num_retries = kwargs.get("num_retries", 100)
        num_retries_call = kwargs.get("num_retries_call", 5)
        self.use_condenser = kwargs.get("use_condenser", False)
        self.use_tor = kwargs.get("use_tor", False)
        self.disable_chain_detection = kwargs.get("disable_chain_detection", False)
        self.known_chains = known_chains
        custom_chain = kwargs.get("custom_chains", {})
        if len(custom_chain) > 0:
            for c in custom_chain:
                if c not in self.known_chains:
                    self.known_chains[c] = custom_chain[c]

        self.nodes = Nodes(urls, num_retries, num_retries_call)
        if self.nodes.working_nodes_count == 0:
            self.current_rpc = self.rpc_methods["offline"]

        self.user = user
        self.password = password
        self.ws = None
        self.url = None
        self.session = None
        self.rpc_queue = []
        if kwargs.get("autoconnect", True):
            self.rpcconnect()

    @property
    def num_retries(self):
        return self.nodes.num_retries

    @property
    def num_retries_call(self):
        return self.nodes.num_retries_call

    @property
    def error_cnt_call(self):
        return self.nodes.error_cnt_call

    @property
    def error_cnt(self):
        return self.nodes.error_cnt

    def get_request_id(self):
        """Get request id."""
        self._request_id += 1
        return self._request_id

    def next(self):
        """Switches to the next node url"""
        if self.ws:
            try:
                self.rpcclose()
            except Exception as e:
                log.warning(str(e))
        self.rpcconnect()

    def is_appbase_ready(self):
        """Check if node is appbase ready"""
        return self.current_rpc in [self.rpc_methods["wsappbase"], self.rpc_methods["appbase"]]

    def get_use_appbase(self):
        """Returns True if appbase ready and appbase calls are set"""
        return not self.use_condenser and self.is_appbase_ready()

    def rpcconnect(self, next_url=True):
        """Connect to next url in a loop."""
        if self.nodes.working_nodes_count == 0:
            return
        while True:
            if next_url:
                self.url = next(self.nodes)
                self.nodes.reset_error_cnt_call()
                log.debug("Trying to connect to node %s" % self.url)
                if self.url[:3] == "wss":
                    self.ws = create_ws_instance(use_ssl=True)
                    self.ws.settimeout(self.timeout)
                    self.current_rpc = self.rpc_methods["wsappbase"]
                elif self.url[:2] == "ws":
                    self.ws = create_ws_instance(use_ssl=False)
                    self.ws.settimeout(self.timeout)
                    self.current_rpc = self.rpc_methods["wsappbase"]
                else:
                    self.ws = None
                    self.session = shared_session_instance()
                    if self.use_tor:
                        self.session.proxies = {}
                        self.session.proxies["http"] = "socks5h://localhost:9050"
                        self.session.proxies["https"] = "socks5h://localhost:9050"
                    self.current_rpc = self.rpc_methods["appbase"]
                    self.headers = {
                        "User-Agent": "nectar v%s" % (nectar_version),
                        "content-type": "application/json; charset=utf-8",
                    }
            try:
                if self.ws:
                    self.ws.connect(self.url)
                    self.rpclogin(self.user, self.password)
                if self.disable_chain_detection:
                    # Set to appbase rpc format
                    if self.current_rpc == self.rpc_methods["ws"]:
                        self.current_rpc = self.rpc_methods["wsappbase"]
                    else:
                        self.current_rpc = self.rpc_methods["appbase"]
                    break
                try:
                    props = None
                    if not self.use_condenser:
                        props = self.get_config(api="database")
                    else:
                        props = self.get_config()
                except Exception as e:
                    if re.search("Bad Cast:Invalid cast from type", str(e)):
                        # retry with not appbase
                        if self.current_rpc == self.rpc_methods["wsappbase"]:
                            self.current_rpc = self.rpc_methods["ws"]
                        else:
                            self.current_rpc = self.rpc_methods["appbase"]
                        props = self.get_config(api="database")
                if props is None:
                    raise RPCError("Could not receive answer for get_config")
                if is_network_appbase_ready(props):
                    if self.ws:
                        self.current_rpc = self.rpc_methods["wsappbase"]
                    else:
                        self.current_rpc = self.rpc_methods["appbase"]
                break
            except KeyboardInterrupt:
                raise
            except Exception as e:
                self.nodes.increase_error_cnt()
                do_sleep = not next_url or (next_url and self.nodes.working_nodes_count == 1)
                self.nodes.sleep_and_check_retries(str(e), sleep=do_sleep)
                next_url = True

    def rpclogin(self, user, password):
        """Login into Websocket"""
        if self.ws and self.current_rpc == self.rpc_methods["ws"] and user and password:
            self.login(user, password, api="login_api")

    def rpcclose(self):
        """Close Websocket"""
        if self.ws is None:
            return
        # if self.ws.connected:
        self.ws.close()

    def request_send(self, payload):
        if self.user is not None and self.password is not None:
            response = self.session.post(
                self.url,
                data=payload,
                headers=self.headers,
                timeout=self.timeout,
                auth=(self.user, self.password),
            )
        else:
            response = self.session.post(
                self.url, data=payload, headers=self.headers, timeout=self.timeout
            )
        if response.status_code == 401:
            raise UnauthorizedError
        return response

    def ws_send(self, payload):
        if self.ws is None:
            raise RPCConnection("No websocket available!")
        self.ws.send(payload)
        reply = self.ws.recv()
        return reply

    def version_string_to_int(self, network_version):
        version_list = network_version.split(".")
        return int(int(version_list[0]) * 1e8 + int(version_list[1]) * 1e4 + int(version_list[2]))

    def get_network(self, props=None):
        """Identify the connected network. This call returns a
        dictionary with keys chain_id, core_symbol and prefix
        """
        if props is None:
            props = self.get_config(api="database")
        chain_id = None
        network_version = None
        blockchain_name = None
        chain_config = None
        prefix = None
        symbols = []
        chain_assets = []

        prefix_count = {}
        for key in props:
            if key.split("_")[0] in prefix_count:
                prefix_count[key.split("_")[0]] += 1
            else:
                prefix_count[key.split("_")[0]] = 1
        if len(prefix_count) > 0:
            sorted_prefix_count = sorted(prefix_count.items(), key=lambda x: x[1], reverse=True)
            if sorted_prefix_count[0][1] > 1:
                blockchain_name = sorted_prefix_count[0][0]
        if blockchain_name is None and "HIVE_CHAIN_ID" in props and "STEEM_CHAIN_ID" in props:
            del props["STEEM_CHAIN_ID"]

        for key in props:
            if key[-8:] == "CHAIN_ID" and blockchain_name is None:
                chain_id = props[key]
                blockchain_name = key.split("_")[0]
            elif key[-8:] == "CHAIN_ID" and key.split("_")[0] == blockchain_name:
                chain_id = props[key]
            elif key[-13:] == "CHAIN_VERSION" and blockchain_name is None:
                network_version = props[key]
            elif key[-13:] == "CHAIN_VERSION" and key.split("_")[0] == blockchain_name:
                network_version = props[key]
            elif key[-14:] == "ADDRESS_PREFIX" and blockchain_name is None:
                prefix = props[key]
            elif key[-14:] == "ADDRESS_PREFIX" and key.split("_")[0] == blockchain_name:
                prefix = props[key]
            elif key[-6:] == "SYMBOL":
                value = {}
                value["asset"] = props[key]["nai"]
                value["precision"] = props[key]["decimals"]
                if (
                    "IS_TEST_NET" in props
                    and props["IS_TEST_NET"]
                    and "nai" in props[key]
                    and props[key]["nai"] == "@@000000013"
                ):
                    value["symbol"] = "TBD"
                elif (
                    "IS_TEST_NET" in props
                    and props["IS_TEST_NET"]
                    and "nai" in props[key]
                    and props[key]["nai"] == "@@000000021"
                ):
                    value["symbol"] = "TESTS"
                else:
                    value["symbol"] = key[:-7]
                value["id"] = -1
                symbols.append(value)
        symbol_id = 0
        if len(symbols) == 2:
            symbol_id = 1
        for s in sorted(symbols, key=lambda self: self["asset"], reverse=False):
            s["id"] = symbol_id
            symbol_id += 1
            chain_assets.append(s)
        if (
            chain_id is not None
            and network_version is not None
            and len(chain_assets) > 0
            and prefix is not None
        ):
            chain_config = {
                "prefix": prefix,
                "chain_id": chain_id,
                "min_version": network_version,
                "chain_assets": chain_assets,
            }

        if chain_id is None:
            raise RPCError("Connecting to unknown network!")
        highest_version_chain = None
        for k, v in list(self.known_chains.items()):
            if (
                blockchain_name is not None
                and blockchain_name not in k
                and blockchain_name != "STEEMIT"
                and blockchain_name != "CHAIN"
            ):
                continue
            if v["chain_id"] == chain_id and self.version_string_to_int(
                v["min_version"]
            ) <= self.version_string_to_int(network_version):
                if highest_version_chain is None:
                    highest_version_chain = v
                elif self.version_string_to_int(v["min_version"]) > self.version_string_to_int(
                    highest_version_chain["min_version"]
                ):
                    highest_version_chain = v
        if highest_version_chain is None and chain_config is not None:
            return chain_config
        elif highest_version_chain is None:
            raise RPCError("Connecting to unknown network!")
        else:
            return highest_version_chain

    def _check_for_server_error(self, reply):
        """Checks for server error message in reply"""
        if re.search("Internal Server Error", reply) or re.search("500", reply):
            raise RPCErrorDoRetry("Internal Server Error")
        elif re.search("Not Implemented", reply) or re.search("501", reply):
            raise RPCError("Not Implemented")
        elif re.search("Bad Gateway", reply) or re.search("502", reply):
            raise RPCErrorDoRetry("Bad Gateway")
        elif re.search("Too Many Requests", reply) or re.search("429", reply):
            raise RPCErrorDoRetry("Too Many Requests")
        elif (
            re.search("Service Temporarily Unavailable", reply)
            or re.search("Service Unavailable", reply)
            or re.search("503", reply)
        ):
            raise RPCErrorDoRetry("Service Temporarily Unavailable")
        elif (
            re.search("Gateway Time-out", reply)
            or re.search("Gateway Timeout", reply)
            or re.search("504", reply)
        ):
            raise RPCErrorDoRetry("Gateway Time-out")
        elif re.search("HTTP Version not supported", reply) or re.search("505", reply):
            raise RPCError("HTTP Version not supported")
        elif re.search("Variant Also Negotiates", reply) or re.search("506", reply):
            raise RPCError("Variant Also Negotiates")
        elif re.search("Insufficient Storage", reply) or re.search("507", reply):
            raise RPCError("Insufficient Storage")
        elif re.search("Loop Detected", reply) or re.search("508", reply):
            raise RPCError("Loop Detected")
        elif re.search("Bandwidth Limit Exceeded", reply) or re.search("509", reply):
            raise RPCError("Bandwidth Limit Exceeded")
        elif re.search("Not Extended", reply) or re.search("510", reply):
            raise RPCError("Not Extended")
        elif re.search("Network Authentication Required", reply) or re.search("511", reply):
            raise RPCError("Network Authentication Required")
        else:
            raise RPCError("Client returned invalid format. Expected JSON!")

    def rpcexec(self, payload):
        """
        Execute a call by sending the payload.

        :param json payload: Payload data
        :raises ValueError: if the server does not respond in proper JSON format
        :raises RPCError: if the server returns an error
        """
        log.debug(f"Payload: {json.dumps(payload)}")
        if self.nodes.working_nodes_count == 0:
            raise WorkingNodeMissing("No working nodes available.")
        if self.url is None:
            raise RPCConnection("RPC is not connected!")

        reply = {}
        response = None
        while True:
            self.nodes.increase_error_cnt_call()
            try:
                if (
                    self.current_rpc == self.rpc_methods["ws"]
                    or self.current_rpc == self.rpc_methods["wsappbase"]
                ):
                    reply = self.ws_send(json.dumps(payload, ensure_ascii=False).encode("utf8"))
                else:
                    response = self.request_send(
                        json.dumps(payload, ensure_ascii=False).encode("utf8")
                    )
                    reply = response.text
                if not bool(reply):
                    try:
                        self.nodes.sleep_and_check_retries("Empty Reply", call_retry=True)
                    except CallRetriesReached:
                        self.nodes.increase_error_cnt()
                        self.nodes.sleep_and_check_retries(
                            "Empty Reply", sleep=False, call_retry=False
                        )
                        self.rpcconnect()
                else:
                    break
            except KeyboardInterrupt:
                raise
            except WebSocketConnectionClosedException as e:
                self.nodes.increase_error_cnt()
                self.nodes.sleep_and_check_retries(str(e), sleep=False, call_retry=False)
                self.rpcconnect()
            except ConnectionError as e:
                self.nodes.increase_error_cnt()
                self.nodes.sleep_and_check_retries(str(e), sleep=False, call_retry=False)
                self.rpcconnect()
            except WebSocketTimeoutException as e:
                self.nodes.increase_error_cnt()
                self.nodes.sleep_and_check_retries(str(e), sleep=False, call_retry=False)
                self.rpcconnect()
            except Exception as e:
                self.nodes.increase_error_cnt()
                self.nodes.sleep_and_check_retries(str(e), sleep=False, call_retry=False)
                self.rpcconnect()

        try:
            if response is None:
                try:
                    ret = json.loads(reply, strict=False)
                except ValueError:
                    log.error(f"Non-JSON response: {reply} Node: {self.url}")
                    self._check_for_server_error(reply)
                    raise RPCError("Invalid response format")
            else:
                ret = response.json()
        except ValueError:
            self._check_for_server_error(reply)

        log.debug(f"Reply: {json.dumps(reply)}")

        if isinstance(ret, dict) and "error" in ret:
            if isinstance(ret["error"], dict):
                error_message = ret["error"].get(
                    "detail", ret["error"].get("message", "Unknown error")
                )
                raise RPCError(error_message)
        elif isinstance(ret, list):
            ret_list = []
            for r in ret:
                if isinstance(r, dict) and "error" in r:
                    error_message = r["error"].get(
                        "detail", r["error"].get("message", "Unknown error")
                    )
                    raise RPCError(error_message)
                elif isinstance(r, dict) and "result" in r:
                    ret_list.append(r["result"])
                else:
                    ret_list.append(r)
            self.nodes.reset_error_cnt_call()
            return ret_list
        elif isinstance(ret, dict) and "result" in ret:
            self.nodes.reset_error_cnt_call()
            return ret["result"]
        else:
            log.error(f"Unexpected response format: {ret} Node: {self.url}")
            raise RPCError(f"Unexpected response format: {ret}")

    # End of Deprecated methods
    ####################################################################
    def __getattr__(self, name):
        """Map all methods to RPC calls and pass through the arguments."""

        def method(*args, **kwargs):
            api_name = get_api_name(self.is_appbase_ready(), *args, **kwargs)
            if self.is_appbase_ready() and self.use_condenser and api_name != "bridge":
                api_name = "condenser_api"
            if api_name is None:
                api_name = "database_api"

            # let's be able to define the num_retries per query
            stored_num_retries_call = self.nodes.num_retries_call
            self.nodes.num_retries_call = kwargs.get("num_retries_call", stored_num_retries_call)
            add_to_queue = kwargs.get("add_to_queue", False)
            query = get_query(
                self.is_appbase_ready() and not self.use_condenser or api_name == "bridge",
                self.get_request_id(),
                api_name,
                name,
                args,
            )
            if add_to_queue:
                self.rpc_queue.append(query)
                self.nodes.num_retries_call = stored_num_retries_call
                return None
            elif len(self.rpc_queue) > 0:
                self.rpc_queue.append(query)
                query = self.rpc_queue
                self.rpc_queue = []
            r = self.rpcexec(query)
            self.nodes.num_retries_call = stored_num_retries_call
            return r

        return method



# This Python file uses the following encoding: utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nectarapi.rpcutils import (
    get_api_name,
    get_query,
    is_network_appbase_ready,
)


class Testcases(unittest.TestCase):
    def test_is_network_appbase_ready(self):
        self.assertTrue(is_network_appbase_ready({"STEEM_BLOCKCHAIN_VERSION": "0.19.10"}))
        self.assertTrue(is_network_appbase_ready({"STEEM_BLOCKCHAIN_VERSION": "0.19.10"}))
        self.assertFalse(is_network_appbase_ready({"STEEMIT_BLOCKCHAIN_VERSION": "0.19.2"}))
        self.assertFalse(is_network_appbase_ready({"STEEMIT_BLOCKCHAIN_VERSION": "0.19.2"}))

    def test_get_api_name(self):
        self.assertEqual(get_api_name(True, api="test"), "test_api")
        self.assertEqual(get_api_name(True, api="test_api"), "test_api")
        self.assertEqual(get_api_name(True, api="jsonrpc"), "jsonrpc")

        self.assertEqual(get_api_name(True), "condenser_api")
        self.assertEqual(get_api_name(False, api="test"), "test_api")
        self.assertEqual(get_api_name(False, api="test_api"), "test_api")
        self.assertTrue(get_api_name(False, api="") is None)
        self.assertTrue(get_api_name(False) is None)

    def test_get_query(self):
        query = get_query(True, 1, "test_api", "test", args="")
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))

        args = ({"a": "b"},)
        query = get_query(True, 1, "test_api", "test", args=args)
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0])

        args = ([{"a": "b"}, {"a": "c"}],)
        query_list = get_query(True, 1, "test_api", "test", args=args)
        query = query_list[0]
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0][0])
        query = query_list[1]
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 2)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0][1])

        args = ("b",)
        query = get_query(True, 1, "test_api", "test", args=args)
        self.assertEqual(query["method"], "call")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], list))
        self.assertEqual(query["params"], ["test_api", "test", ["b"]])

        args = ("b",)
        query = get_query(True, 1, "condenser_api", "test", args=args)
        self.assertEqual(query["method"], "call")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], list))
        self.assertEqual(query["params"], ["condenser_api", "test", ["b"]])

        args = ("b",)
        query = get_query(False, 1, "test_api", "test", args=args)
        self.assertEqual(query["method"], "call")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], list))
        self.assertEqual(query["params"], ["test_api", "test", ["b"]])
